/* $Id: CodeGen.cpp,v 1.20 2003/12/12 09:47:26 markus Exp $ */

/*
 * Copyright (c) 2003 Markus Hennecke
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the Author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Implementation of the codegenerator.
 */

// system includes:
#include <cctype>
#include <cstdlib>

#ifdef DEBUG
#include <iostream>
#endif

// program includes:
#include "CodeGen.hpp"
#include "global_options.hpp"

namespace CodeGenerator {

  /* The table with the meta-instructions */
  static const OPCODE_t metas[] = {
    { "ORG", 0x0, Lexer::ARG_Relative },
    { "EQU", 0x0, Lexer::ARG_Relative },
    { "EQUIVALENT", 0x0, Lexer::ARG_Relative },
    { "LIST", 0x0, Lexer::ARG_None },
    { "NOLIST", 0x0, Lexer::ARG_None },
    { "WORD", 0x0, Lexer::ARG_Address | Lexer::ARG_Operator },
    { "TEXT", 0x0, Lexer::ARG_Quote },
    { NULL, 0x0, 0 }
  };

  /* The table with the instructions */
  static const OPCODE_t opcodes[] = {
    { "HLT", 0x0, Lexer::ARG_None },
    { "LAD", 0x1, Lexer::ARG_Address | Lexer::ARG_Operator | 
      Lexer::ARG_Relative | Lexer::ARG_Indirect | Lexer::ARG_Index },
    { "SPI", 0x2, Lexer::ARG_Address | Lexer::ARG_Relative |
      Lexer::ARG_Indirect | Lexer::ARG_Index },
    { "ADD", 0x3, Lexer::ARG_Address | Lexer::ARG_Operator | 
      Lexer::ARG_Relative | Lexer::ARG_Indirect | Lexer::ARG_Index },
    { "SUB", 0x4, Lexer::ARG_Address | Lexer::ARG_Operator | 
      Lexer::ARG_Relative | Lexer::ARG_Indirect | Lexer::ARG_Index },
    { "MUL", 0x5, Lexer::ARG_Address | Lexer::ARG_Operator | 
      Lexer::ARG_Relative | Lexer::ARG_Indirect | Lexer::ARG_Index },
    { "DIV", 0x6, Lexer::ARG_Address | Lexer::ARG_Operator | 
      Lexer::ARG_Relative | Lexer::ARG_Indirect | Lexer::ARG_Index },
    { "JMP", 0x7, Lexer::ARG_Address | Lexer::ARG_Relative | 
      Lexer::ARG_Indirect | Lexer::ARG_Index },
    { "JEZ", 0x8, Lexer::ARG_Address | Lexer::ARG_Relative |
      Lexer::ARG_Indirect | Lexer::ARG_Index },
    { "JGZ", 0x9, Lexer::ARG_Address | Lexer::ARG_Relative |
      Lexer::ARG_Indirect | Lexer::ARG_Index },
    { "JLZ", 0xa, Lexer::ARG_Address | Lexer::ARG_Relative |
      Lexer::ARG_Indirect | Lexer::ARG_Index },
    { "SWAP", 0xb, Lexer::ARG_None },
    { "PUSHAC", 0xc1, Lexer::ARG_None },
    { "POPAC", 0xc2, Lexer::ARG_None },
    { "ADDST", 0xc3, Lexer::ARG_None },
    { "SUBST", 0xc4, Lexer::ARG_None },
    { "MULST", 0xc5, Lexer::ARG_None },
    { "DIVST", 0xc6, Lexer::ARG_None },
    { "LSA", 0xc7, Lexer::ARG_None },
    { "LDS", 0xc8, Lexer::ARG_None },
    { "PUSH", 0xcb, Lexer::ARG_Operator },
    { "POP", 0xcc, Lexer::ARG_Address },
    { "JSR", 0xd, Lexer::ARG_Address | Lexer::ARG_Relative |
      Lexer::ARG_Indirect | Lexer::ARG_Index },
    { "RET", 0xe , Lexer::ARG_None },
    { "CLI", 0xf , Lexer::ARG_None },
    { "LDI", 0xf1, Lexer::ARG_None },
    { "LIA", 0xf2, Lexer::ARG_None },
    { "INK", 0xf3, Lexer::ARG_None },
    { "DEK", 0xf4, Lexer::ARG_None },
    { "LDX", 0xf7, Lexer::ARG_Address },
    { NULL, 0x0, Lexer::ARG_None }
  };

  // CodeGen::~CodeGen()
  // Virtual destructor. Nothing to do here.
  CodeGen::~CodeGen()
  { }

  // const OPCODE_t *CodeGen::lookupOpcode(const std::string &)
  // Returns a pointer to the instruction-definition if the opcode
  // is listed in the table. If it is not in the table NULL is
  // returned.
  const OPCODE_t *CodeGen::lookupOpcode(const std::string &s)
  {
    for (int i = 0; opcodes[i].mnemonic != NULL; i++) {
      if (opcodes[i].mnemonic == s) 
        return &opcodes[i];
    }
    return NULL;
  }

  // bool CodeGen::lookupMeta(const std::string &)
  // Returns true if the meta-instruction is listed in the table. If
  // it is not in the table false is returned.
  bool CodeGen::lookupMeta(const std::string &meta)
  {
    for (int i = 0; metas[i].mnemonic != NULL; i++) {
      if (metas[i].mnemonic == meta) 
        return true;
    }
    return false;
  }

  // inline static bool is_number(const std::string &)
  // Helper function to test a string for a possible number.
  inline static bool is_number(const std::string &s)
  {
    return ((isdigit(s[0]) != 0) || (s[0] == '$'));
  }

  // bool CodeGen::doMeta(const std::string &, const std::string &,
  //                      const std::string &, Lexer::ARG_TYPE_t,
  //                      SymbolTable::SymbolTable &ST, const int,
  //                      const bool)
  // Handles a meta-instruction. Throws syntax_error if an error
  // occurs.
  bool CodeGen::doMeta(const std::string &label, const std::string &meta,
                       const std::string &arg, Lexer::ARG_TYPE_t type,
                       SymbolTable::SymbolTable &ST, const int pass,
                       const bool negative)
    throw (Error::syntax_error, Error::internal_error)
  {
    if (lookupMeta(meta)) {

      if (meta == "ORG") {
        if (::listfile && ::listfileallowed) {
          SH.list_write_ascii(" ", 12);
        }
        // has to look like ORG, xxxxx where xxxx is a number or a label
        // the label has to be defined prior using ORG
        if ((type == Lexer::ARG_Relative) || negative) {
          if (is_number(arg)) {
            IP = atoi(arg.c_str());
            SH.list_set_ip(IP);
            return true;
          } else {
            // We habe a label
            const SymbolTable::Symbol *sym = ST.lookup(arg);
            if (! sym->is_defined())
              throw Error::syntax_error("Undefined Label", SH.getLineNr());
            IP = sym->getValue();
            SH.list_set_ip(IP);
            return true;
          }
        }
        throw Error::syntax_error("Invalid Argument", SH.getLineNr());

      } else if (meta == "EQU" || meta == "EQUIVALENT") {
        if (type == Lexer::ARG_Relative) {

          if (::listfile && ::listfileallowed) {
            SH.list_write_ascii(" ", 12);
          }
          // handle the label that is to be defined
          const SymbolTable::Symbol *lbl = ST.lookup(label);
          if (lbl) {
            if (pass == 1)
              throw Error::syntax_error("Redefined label", SH.getLineNr());
          } else {
            if (pass > 1)
              throw Error::internal_error(SH.getLineNr());
          }

          // if we get here we can handle the argument
          bool defined = true;
          int value;
          if (is_number(arg)) {
            value = atoi(arg.c_str());
          } else {
            const SymbolTable::Symbol *argument = ST.lookup(arg);
            if (argument) {
              if (argument->is_defined()) {
                value = argument->getValue();
              } else {
                if (pass > 1)
                  throw Error::syntax_error("Undefined symbol", SH.getLineNr());
                value = 0;
                defined = false;
              }
            } else {
              if (pass > 1) 
                throw Error::syntax_error("Undefined symbol", SH.getLineNr());
              // pass 1 we will guess the value
              value = 0;
              defined = false;
            }
          }
          SymbolTable::Symbol sym(label, defined, value);
          ST.change(sym);

          return true;
        }
        throw Error::syntax_error("Invalid Argument", SH.getLineNr());

      } else if (meta == "WORD") {
        if (type == Lexer::ARG_Operator) {
          // it is a number or a relative address
          if (! is_number(arg)) {
            if (pass > 1) {
              const SymbolTable::Symbol *sym = ST.lookup(arg);
              if (! sym->is_defined())
                throw Error::syntax_error("Undefined Label", SH.getLineNr());
              emit_data(sym->getValue());
            }
          } else if (pass > 1) {
            if (negative)
              emit_data(-atoi(arg.c_str()));
            else 
              emit_data(atoi(arg.c_str()));
          }
        } else if (type == Lexer::ARG_Address) {
          if (! is_number(arg)) {
            // it is a label
            if (pass > 1) {
              const SymbolTable::Symbol *sym = ST.lookup(arg);
              if (! sym->is_defined())
                throw Error::syntax_error("Undefined Label", SH.getLineNr());
              emit_data(sym->getValue());
            }
          } else {
            if (pass > 1) {
              if (negative)
                emit_data(-atoi(arg.c_str()));
              else 
                emit_data(atoi(arg.c_str()));
            }
          }
        } else 
          throw Error::syntax_error("Wrong address mode for WORD instruction.", SH.getLineNr());
        IP++;
        SH.list_set_ip(IP);
        return true;

      } else if (meta == "TEXT") {
        if (type == Lexer::ARG_Quote) {
          if (pass > 1)
            emit_string(arg);
          IP += arg.length();
          SH.list_set_ip(IP);
          return true;
        } else {
          throw Error::syntax_error("No text as argument.", SH.getLineNr());
        }
      } else if (meta == "LIST") {
        if (type == Lexer::ARG_None) {
          ::listfileallowed = true;
          return true;
        } else {
          throw Error::syntax_error("No argument to LIST allowed.", SH.getLineNr());
        }
      } else if (meta == "NOLIST") {
        if (type == Lexer::ARG_None) {
          ::listfileallowed = false;
          return true;
        } else {
          throw Error::syntax_error("No argument to LIST allowed.", SH.getLineNr());
        }
      }
    }
    return false;
  }

  // bool CodeGen::generate(const std::string &op, const std::string &arg,
  //                        const Lexer::ARG_TYPE_t mode,
  //                        SymbolTable::SymbolTable &table,
  //                        const int pass, const bool lastpass)
  //   throw (Error::syntax_error, Error::internal_error)
  // Generates code according to the op, arg and the mode. If an error
  // occurs exceptions are thrown. If the opcode is not known false
  // is returned.
  bool CodeGen::generate(const std::string &op, const std::string &arg,
                         const Lexer::ARG_TYPE_t mode,
                         SymbolTable::SymbolTable &table,
                         const int pass, const bool lastpass,
                         const bool negative)
    throw (Error::syntax_error, Error::internal_error)
  {
    const OPCODE_t *opc = lookupOpcode(op);

    // unknown Opcode?
    if (! opc) {
      return false;
    }

    if (! (opc->modes & mode)) {
      throw Error::syntax_error("Invalid address mode", SH.getLineNr());
    }

    int num, first = 0, second = 0, value = 0;
    switch (mode) {
    case Lexer::ARG_None:
      num = 1;
      first = (opc->opcode << ((opc->opcode < 0x10) ? 12 : 8));
      break;

    case Lexer::ARG_Address:
      //if (arg == "")
      //  throw Error::syntax_error("No valid argument", SH.getLineNr());
      if (! is_number(arg)) {
        // it is a label
        const SymbolTable::Symbol *sym = table.lookup(arg);
        if (sym) {
          value = sym->getValue();
          if (sym->is_defined()) {
            num = (value < 1024) ? 1 : 2;
          } else {
            // assuming 2 words if it is not already defined
            num = 2;
          }
        } else {
          if (pass > 1) {
            std::string errorstr = "Undefined symbol: " + arg;
            
            throw Error::syntax_error(errorstr, SH.getLineNr());
          } else {
            // give it a usefull value to not provoke out of range errors
            // on the first pass
            value = 0;
          }
          num = 2;
        }
      } else {
        value = atoi(arg.c_str());
        num = (value < 1024) ? 1 : 2;
      }

      if (num == 2) {
        second = value;
        first = 0x0800; // set the D bit
      } else {
        first = value;
      }

      // we have to handle ldx special
      if (opc->opcode == 0xf7) {
        second = value;
        first = 0xff00;
        num = 2;
      } else {
        first |= (opc->opcode << 12);
      }

      if (value < 0)
        throw Error::syntax_error("Address may not be negative",
                                        SH.getLineNr());
      if (value > 65535)
        throw Error::syntax_error("Argument too large", SH.getLineNr());

      break;
      
    case Lexer::ARG_Operator:
      if (! is_number(arg)) {
        // it is a symbolic
        const SymbolTable::Symbol *sym = table.lookup(arg);
        if (sym) {
          value = sym->getValue();
          if (negative)
            value = -value;
          if (sym->is_defined()) {
            num = ((value >= -128) && (value <= 127)) ? 1 : 2;
          } else {
            // assuming 2 words if it is not already defined
            num = 2;
          }
        } else {
          if (pass > 1) {
            std::string errorstr = "Undefined symbol: " + arg;
            
            throw Error::syntax_error(errorstr, SH.getLineNr());
          } else {
            // give it a usefull value to not provoke out of range errors
            // on the first pass
            value = 0;
          }
          num = 2;
        }
      } else {
        if (negative)
          value = -atoi(arg.c_str());
        else
          value = atoi(arg.c_str());
        num = ((value >= -128) && (value <= 127)) ? 1 : 2;
      }

      // Check for 2 word stack instructions, this is "push" only
      if (opc->opcode == 0xcb) {
        num = 2;
      }

      if (num == 2) {
        second = value & 0x0ffff;
        if (opc->opcode != 0xcb) {
          first = 0x0c00; // set the D bit
        }
      } else {
        // if it is negative we have to keep the last 8 bits only
        value &= 0x00ff;
        first = value | 0x0400;
      }
      
      // shift 12 bits normaly, but not on opcodes having 8 bits predefined
      if (opc->opcode > 0xf) {
        first |= (opc->opcode << 8);
      } else {
        first |= (opc->opcode << 12);
      }

      // Check if we are in 16bit range...
      // Treat values from 32767..65535 as negatives but accept them.
      if ((value < -32768) || (value > 65535)) {
        std::cout << value << std::endl;
        throw Error::syntax_error("Argument out of range", SH.getLineNr());
      }
      break;

    case Lexer::ARG_Relative:
      if (! is_number(arg)) {
        // lookup the symbol
        const SymbolTable::Symbol *sym = table.lookup(arg);
        if (sym) {
          value = sym->getValue() - IP;
          if (sym->is_defined()) {
            num = ((value >= -128) && (value <= 127)) ? 1 : 2;
          } else {
            num = 2;
          }
        } else {
          if (pass > 1) {
            std::string errorstr = "Undefined symbol: " + arg;
            
            throw Error::syntax_error(errorstr, SH.getLineNr());
          } else {
            // give it a usefull value to not provoke out of range errors
            // on the first pass
            value = 0;
          }
          num = 2;
        }
      } else {
        value = atoi(arg.c_str()) - IP;
        num = ((value >= -128) && (value <= 127)) ? 1 : 2;
      }

      if (num == 2) {
        second = value;
        first = 0x0d00;
      } else {
        value &= 0x00ff;
        first = value | 0x0500;
      }
      first |= (opc->opcode << 12);

      // Check if we are in 16bit range...
      if ((value < -32768) || (value > 32767)) 
        throw Error::syntax_error("Argument out of range", SH.getLineNr());

      break;

    case Lexer::ARG_Indirect:
      if (! is_number(arg)) {
        // lookup the symbol
        const SymbolTable::Symbol *sym = table.lookup(arg);
        if (sym) {
          value = sym->getValue();
          if (sym->is_defined()) {
            num = (value < 256) ? 1 : 2;
          } else {
            num = 2;
          }
        } else {
          if (pass > 1) {
            std::string errorstr = "Undefined symbol: " + arg;
            
            throw Error::syntax_error(errorstr, SH.getLineNr());
          } else {
            // give it a usefull value to not provoke out of range errors
            // on the first pass
            value = 0;
          }
          num = 2;
        }
      } else {
        value = atoi(arg.c_str());
        num = (value < 256) ? 1 : 2;
      }

      if (num == 2) {
        second = value;
        first = 0x0e00;
      } else {
        first = value | 0x0600;
      }
      first |= (opc->opcode << 12);

      // Check for a valid 16bit address
      if (value < 0)
        throw Error::syntax_error("Address may not be negative",
                                        SH.getLineNr());
      if (value > 65535)
        throw Error::syntax_error("Argument too large", SH.getLineNr());

      break;

    case Lexer::ARG_Index:
      if (! is_number(arg)) {
        // lookup the symbol
        const SymbolTable::Symbol *sym = table.lookup(arg);
        if (sym) {
          value = sym->getValue();
          if (sym->is_defined()) {
            num = ((value >= -128) && (value <= 127)) ? 1 : 2;
          } else {
            num = 2;
          }
        } else {
          if (pass > 1) {
            std::string errorstr = "Undefined symbol: " + arg;
            
            throw Error::syntax_error(errorstr, SH.getLineNr());
          } else {
            // give it a usefull value to not provoke out of range errors
            // on the first pass
            value = 0;
          }
          num = 2;
        }
      } else {
        if (negative)
          value = -atoi(arg.c_str());
        else
          value = atoi(arg.c_str());
        num = ((value >= -128) && (value <= 127)) ? 1 : 2;
      }

      if (num == 2) {
        second = value;
        first = 0x0f00;
      } else {
        value &= 0x00ff;
        first = value | 0x0700;
      }
      first |= (opc->opcode << 12);

      // Check if we are in 16bit range...
      if ((value < -32768) || (value > 32767)) 
        throw Error::syntax_error("Argument out of range", SH.getLineNr());

      break;

    default:
      throw Error::internal_error(SH.getLineNr());
    }

    // store the length of the assembled output in opcode_lengths
    if (!lastpass) {
      opcode_lengths.push(num);
    } else {
      num = opcode_lengths.front(); 
      if (num == 2) {
        // set the D bit and set the second to value
        first &= (mode == Lexer::ARG_Address) ? 0xf000 : 0xf700;
        first |= 0x0800;
        second = value;
      }
    }
    if (pass > 1) {
      // remove the first length-object from the queue
      opcode_lengths.pop();

      // emit the code
      emit_code(first, second);

#ifdef DEBUG     
      SH.obj_write_ascii(";", 2);
      SH.obj_write_ascii(op, 8);
      if (negative)
        SH.obj_write_ascii("-", 1);
      SH.obj_write_ascii(arg, 1);
      SH.obj_write_ascii("\n", 1);
#endif
    }

    IP += num;
    SH.list_set_ip(IP);
    return true;
  }

  // void CodeGen::emit_code(const int, const int) const
  // Writes the opcode and the optional data to the objfile.
  void CodeGen::emit_code(const int code, const int data) const
  {
    SH.obj_write_hex(IP, 4);
    SH.obj_write_ascii(": ", 2);

    // Test if it is a special one byte instruction like LDS
    // If it is a regulare 2 byte instruction emit two bytes
    if ((code & 0x0800) && ((code & 0xff00) != 0xc800)) {
      SH.obj_write_hex(code, 5);
      SH.obj_write_hex(data, 5);
      if (::listfile && ::listfileallowed) {
        SH.list_write_hex(code, 5);
        SH.list_write_hex(data, 7);
      }
    } else {
      SH.obj_write_hex(code, 10);
      if (::listfile && ::listfileallowed) {
        SH.list_write_hex(code, 12);
      }
    }
#ifndef DEBUG
    SH.obj_write_ascii("\n", 1);
#endif
  }

  // void CodeGen::emit_data(const int) const
  // Puts a word to the objectfile.
  void CodeGen::emit_data(const int data) const
  {
    SH.obj_write_hex(IP, 4);
    SH.obj_write_ascii(": ", 2);

    SH.obj_write_hex(data & 0xffff, 4);
    SH.obj_write_ascii("\n", 1);
    if (::listfile && ::listfileallowed) {
      SH.list_write_hex(data & 0xffff, 12);
    }
  }

  // void CodeGen::emit_string(const std::string &) const
  // Puts a string to the objectfile. This is done by converting every
  // character in the string to a word and writing this word to the file.
  void CodeGen::emit_string(const std::string &s) const
  {
    SH.obj_write_hex(IP, 4);
    SH.obj_write_ascii(": ", 2);

    std::string::const_iterator it;
    for (it = s.begin(); it != s.end(); it++) {
      SH.obj_write_hex(static_cast<int>(*it), 5);
    }
#ifdef DEBUG
    SH.obj_write_ascii("; TEXT \"", 2);
    SH.obj_write_ascii(s, 1);
    SH.obj_write_ascii("\"", 1);
#endif
    SH.obj_write_ascii("\n", 1);

    if (::listfile && ::listfileallowed) {
      it = s.begin();
      if (it != s.end())
        SH.list_write_hex(static_cast<int>(*it), 5);
      SH.list_write_ascii("....", 7);
    }
  }

  // int CodeGen::atoi(const std::string &) const
  // Replacement for ::atoi to allow hex numbers.
  int CodeGen::atoi(const std::string &num_string) const
  {
    std::string s(num_string);
    bool hex = false;

    if (s[0] == '$') {
      hex = true;
      s.erase(0, 1);
    } else {
      // ok, it is decimal...
      return ::atoi(s.c_str());
    }

    // only if we got a hex number
    int num = 0;
    while (s.length() > 0) {
      num *= 16;
      int digit = toupper(s[0]) - 0x30;
      if (digit > 9)
	digit -= 0x7;
      num += digit;
      s.erase(0, 1);
    }
    return num;
  }

}
