/* $Id: Lexer.cpp,v 1.12 2003/12/12 09:47:26 markus Exp $ */

/*
 * Copyright (c) 2003 Markus Hennecke
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the Author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * The implementation of the LexicalAnalyzer class.
 */

// system includes:
#include <cctype>
#include <iostream>

// program includes:
#include "Lexer.hpp"

namespace Lexer {

  // bool LexicalAnalyzer::getLabel(string &)
  // Reads a label from the current line position. Normaly this would be the
  // beginning of the line without any whitespace before the label.
  // A label consists of a lowercase character followed by zero or more
  // characters from the set [0-9A-Za-z_].
  bool LexicalAnalyzer::getLabel(std::string &label)
  {
    label = "";
    if (islower(SH.getchar())) {
      // labels start with a lower case character
      do {
        token.append(1, static_cast<char>(SH.getchar()));
        SH.nextchar();
      } while ((isalnum(SH.getchar())) || ('_' == SH.getchar()));
      // and are finished with a colon
      if (':' == SH.getchar()) {
        SH.nextchar();
        label = token;
        token = "";
        return true;
      }
    }
    return false;
  }

  // bool LexicalAnalyzer::ReadNumber(string &)
  // Reads a number from the current line position
  bool LexicalAnalyzer::ReadNumber(std::string &num)
  {
    bool hex = false;

    // decimal numbers consist of digits only...
    if (SH.getchar() == '$') {
      num.append(1, static_cast<char>(SH.getchar()));
      if (SH.endline())
	return false;
      SH.nextchar();
      hex = true;
    }

    while (hex ? isxdigit(SH.getchar()) : isdigit(SH.getchar())) {
      num.append(1, static_cast<char>(SH.getchar()));
      if (SH.endline())
	break;
      SH.nextchar();
    } 
    return (num.length() > ((hex) ? 1U : 0));
  }

  // bool LexicalAnalyzer::ReadID(string &)
  // Reads a id from the current line position.
  bool LexicalAnalyzer::ReadID(std::string &id)
  {
    // every id, like a label starts with a lowercase character followed
    // by [A-Za-z0-9_]
    if (islower(SH.getchar())) {
      do {
        id.append(1, static_cast<char>(SH.getchar()));
        if (SH.endline())
          break;
        SH.nextchar();
      } while ((isalnum(SH.getchar())) || ('_' == SH.getchar()));
      return true;
    }
    return false;
  }

  // bool LexicalAnalyzer::getArg(string &, enum ARG_TYPE &)
  // Looks for an argument. The argument is preceded by a comma.
  // There are Addresses (@.....), Numbers ([0-9]+ | #id | #[0-9]+),
  // or strings (".*").
  bool LexicalAnalyzer::getArg(std::string &arg, ARG_TYPE_t &type,
                               bool &negative)
  {
    arg = "";
    negative = false;
    // if it is a ',' then we have an argument
    if (',' == SH.getchar()) {
      if (! SH.endline()) {
        SH.nextchar();
      } else {
        // looks like an error
        type = ARG_Unknown;
        return false;
      }
      eatBlanks();

      switch (char c = static_cast<char>(SH.getchar())) {

      case '#':
        // Direct with operator
        type = ARG_Operator;
        goto neg_check;

      case '%':
        // Relative to the index-register
        type = ARG_Index;
      neg_check:
        if (! SH.endline()) {
          SH.nextchar();
        } else {
          // looks like an error
          type = ARG_Unknown;
          return false;
        }
        if (static_cast<char>(SH.getchar()) == '-') {
          // this is negative
          negative = true;
          if (! SH.endline()) {
            SH.nextchar();
          } else {
            // looks like an error
            type = ARG_Unknown;
            return false;
          }
        }
        goto next;

      case '@':
        // Direct with address
        type = ARG_Address;
        if (! SH.endline()) {
          SH.nextchar();
        } else {
          // looks like an error
          type = ARG_Unknown;
          return false;
        }
      next:
        if (! ReadID(arg)) // no id? then it is a number
          if (! ReadNumber(arg)) {
            type = ARG_Unknown; // oops
            return false;
          }
        break;

      case '0': case '1': case '2': case '3':
      case '4': case '5': case '6': case '7':
      case '8': case '9': case '$':
        // Relative to Instruction Pointer
        type = ARG_Relative;
        ReadNumber(arg);
        break;

      case '(':
        // Indirect via another address
        if (! SH.endline()) {
          SH.nextchar();
          if (! ReadID(arg))
            ReadNumber(arg);
          if (static_cast<char>(SH.getchar()) == ')') {
            // if we are at the end of line we won't need another char
            if (! SH.endline())
              SH.nextchar();
            type = ARG_Indirect;
            return true;
          }
        }

        type = ARG_Unknown;
        return false;

        break;

      case '"':
        // It's a string
        if (getQwchar(arg)) {
          type = ARG_Quote;
          return true;
        }
        break;

      default:
        // Could be a symbol...
        if (isalpha(c) && islower(c)) {
          ReadID(arg);
          type = ARG_Relative;
        } else {
          // Hm, somethings wrong in the sourcecode...
          type = ARG_Unknown;
          return false;
        }
      }
    } else {
      type = ARG_None;
    }
    return true;
  }

  // bool LexicalAnalyzer::get_opcode(string &)
  // Reads and returns the opcode in the string argument. An opcode consists
  // of uppercase characters only.
  bool LexicalAnalyzer::getOpcode(std::string &opcode)
  {
    // any opcode consists of uppercase characters
    opcode = "";
    if (! SH.endline()) {
      if (isupper(SH.getchar())) {
        // opcodes are in upper case
        do {
          token.append(1, static_cast<char>(SH.getchar()));
          if (SH.endline())
            break;
        } while (isupper(SH.nextchar()));
        opcode = token;
        token = "";
        // no check for a valid opcode yet
        return true;
      }
    }
    return false;
  }
  
  // void LexicalAnalyzer::eatBlanks()
  // Remove any blanks following the actual position until an other character
  // than blank or the end of line is reached.
  void LexicalAnalyzer::eatBlanks()
  {
    while (! SH.endline()) {
      if (! isspace(SH.getchar()))
        return;
      SH.nextchar();
    }
  }

  // bool LexicalAnalyzer::getComment()
  // Reads from the current line position to the end of line if there is a
  // ';' at the current possition.
  bool LexicalAnalyzer::getComment()
  {
    if (SH.getchar() == ';') {
      while (!SH.endline()) {
        SH.nextchar();
      }
      return true;
    }
    return false;
  }

  // bool LexicalAnalyzer::get_qwchar(string &)
  // Reads the string inside '"'s. If an error occurs false is returned.
  bool LexicalAnalyzer::getQwchar(std::string &qw)
  {
    qw = "";
    if (SH.getchar() == '"') {
      while (!SH.endline()) {
        char ch = SH.nextchar();
        switch (ch) {
        case '\\':
          if (SH.endline()) {
            return false;
          } else {
            ch = SH.nextchar();
          }
        default:
          token.append(1, ch);
          break;
        case '"':
          // we won't need the next character if we are already at eol
          if (! SH.endline())
            SH.nextchar();
          qw = token;
          token = "";
          return true;
        }
      }
    }
    return false;
  }

}
