/* $Id: Parser.cpp,v 1.12 2004/11/21 18:52:51 markus Exp $ */

/*
 * Copyright (c) 2003 Markus Hennecke
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the Author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Implementation of the parser.
 */

// system includes:
#include <string>

// program includes:
#include "Parser.hpp"
#include "global_options.hpp"

namespace Parser {

  //   Parser::Parser(Sourcehandler::SourceHandler &SH,
  //                  Lexer::LexicalAnalyzer &LA,
  //                  CodeGenerator::CodeGen &GEN,
  //                  SymbolTable::SymbolTable &ST)
  // Initializes the references the parser depends on.
  Parser::Parser(Sourcehandler::SourceHandler &SH, Lexer::LexicalAnalyzer &LA,
           CodeGenerator::CodeGen &GEN, SymbolTable::SymbolTable &ST)
    : SH(SH), LA(LA), GEN(GEN), ST(ST)
  { }

  // void Parser::rewind()
  // Rewinds all open files.
  void Parser::rewind()
  {
    SH.rewind();
    GEN.rewind();
  }

  // bool Parser::parse(const int pass, const bool lastpass)
  // Returns true if another pass is needed.
  bool Parser::parse(const int pass, const bool lastpass)
    throw (Error::syntax_error, Error::internal_error)
  {
    bool another_pass = false;

    while (SH.nextchar() != EOF) {
      std::string line, label, opcode, arg;
      Lexer::ARG_TYPE_t type;
      bool negative;

      // Read the whole line and split it in to it's components
      LA.getLabel(label);
      LA.eatBlanks();
      LA.getOpcode(opcode);
      LA.getArg(arg, type, negative);
      if (type == Lexer::ARG_Unknown)
        throw Error::syntax_error("Invalid argument", SH.getLineNr());
      LA.eatBlanks();
      LA.getComment();
      if (! SH.endline()) {
        // Syntax error, no valid line
        // TODO: investigate the error??
        throw Error::syntax_error("No valid line", SH.getLineNr());
      }
      
      // Handle the label if it is there
      if (label != "") {
        const SymbolTable::Symbol *sym = ST.lookup(label);

        if (pass > 1) {
          // if we are in pass>1 and it is not yet defined we have a problem
          if (! sym) {
            // Internal error
            throw Error::internal_error(SH.getLineNr());
          }
        } else {
          // if it is already defined in the first pass raise an exception
          if (sym)
            throw Error::syntax_error("Redefined label", SH.getLineNr());
        }

        // look if it is a user defined value
        // this will be handled by doMeta later in the code
        if ((opcode != "EQU") && (opcode != "EQUIVALENT")) {

          if (pass > 1) {
            if (! lastpass) {
              if (sym->getValue() != GEN.getIP()) {
                // We have to do this again...
                another_pass = true;
                SymbolTable::Symbol symnew(*sym);
                symnew.define(GEN.getIP());
                ST.change(symnew);
              }            
            }
          } else {
            SymbolTable::Symbol sym(label, true, GEN.getIP());
            ST.change(sym);
          }
        }
      }

      // Handle opcode and arg if they are there
      if (opcode != "") {
        // Look if the opcode is valid
        if (GEN.lookupOpcode(opcode)) {
          // generate the code...
          GEN.generate(opcode, arg, type, ST, pass, lastpass, negative);
        } else {
          // look for a meta-code
          if (GEN.lookupMeta(opcode)) {
            if (! GEN.doMeta(label, opcode, arg, type, ST, pass, negative)) {
              // Syntax error
              throw Error::syntax_error("Invalid Meta Instruction",
                                        SH.getLineNr());
            }
          } else {
            // The opcode was not valid
            throw Error::syntax_error("Invalid Opcode", SH.getLineNr());
          }
        }
      } else {
        if (arg != "") {
          // Syntax error: Arg without an opcode
          throw Error::syntax_error("No Opcode", SH.getLineNr());
        }
        
        if (::listfile && ::listfileallowed) {
          SH.list_write_ascii(" ", 12);
        }
      }
     
      if (::listfile && ::listfileallowed) {
        SH.list_write_ascii(SH.getLine(), 1);
      }
    }
    return another_pass;
  }

}
