/* $Id: SourceHandler.cpp,v 1.8 2003/12/12 09:47:26 markus Exp $ */

/*
 * Copyright (c) 2003 Markus Hennecke
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the Author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Class implementation of the source handler. Reads the source file and 
 * provides functions to write a list file.
 */

// program includes:
#include "Error.hpp"
#include "SourceHandler.hpp"
#include "global_options.hpp"

// system includes:
#include <string>
#include <sstream>
#include <fstream>
#include <iostream>

namespace Sourcehandler {

  //  SourceHandler(const std::string &sourcename, const std::string &listname,
  //                const std::string &objname)
  //    throw(Error::file_open_error)
  // Initializes the source handler. Source and listfiles are opened and an
  // exception is thrown if something went wrong.
  // If either the sourcename or the objname is equal to "-" the SourceHandler
  // uses cin eg. cout for its in- and output.
  SourceHandler::SourceHandler(const std::string &sourcename,
                               const std::string &listname,
                               const std::string &objname)
    throw(Error::file_open_error)
    : srcname(sourcename), lstname(listname), objname(objname)
  {
    lst = NULL;
    obj = NULL;
    listfile_ip = 0;

    // create the source file stream and check if we could open it
    src = new std::ifstream(sourcename.c_str());
    if (! src->is_open()) 
      throw Error::file_open_error(sourcename);

    pos = 0;
    linelength = 0;
    linenr = 1;

    // create the list file stream and check if we could open it
    if (::listfile) {
      lst = new std::ofstream(listname.c_str());
      if (! src->is_open())
        throw Error::file_open_error(listname);
    }

    // create the object file stream and check if we could open it
    // if "-" is the object name we will use stdout
    if (objname == "-") {
      obj_output = &std::cout;
      obj_is_stdout = true;
      obj = NULL;
    } else {
      obj_is_stdout = false;
      obj = new std::ofstream(objname.c_str());
      if (! obj->is_open()) 
        throw Error::file_open_error(objname);
      obj_output = obj;
    }
  }

  // SourceHandler::~SourceHandler()
  // Closes and deletes the streams assosiated with the source and the list
  // file.
  SourceHandler::~SourceHandler()
  {
    if (src) {
      src->close();
      delete src;
    }

    if (lst) {
      lst->close();
      delete lst;
    }

    if (! obj_is_stdout) {
      if (obj) {
        obj->close();
        delete obj;
      }
    }
  }

  // void SourceHandler::rewind()  throw (Error::file_open_error)
  // Rewinds the source file and reopens the output file.
  // The listfile-buffer and the objectfile-buffer are flushed.
  void SourceHandler::rewind()
  {
    src->clear();
    src->seekg(0, std::ios::beg);

    pos = 0;
    linelength = 0;
    linenr = 1;

    listfile.clear();
    listfile_ip = 0;
    objfile.clear();
  }

  // int SourceHandler::nextchar()
  // Returns the next character from the source file or EOF if the end of
  // the source file is reached.
  int SourceHandler::nextchar() throw(Error::file_read_error)
  {
    if (pos == linelength) {
      line = "";
      linelength = 0;
      pos = 0;

    listwrite:
      if (::listfile && ::listfileallowed) {
        std::ostringstream ost;
        ost.width(4);
        ost.fill(' ');
        ost << std::endl << linenr;
        list_write_ascii(ost.str(), 6);
        ost.str("");
        ost.setf(std::ios::basefield, std::ios::hex);
        ost.unsetf(std::ios::dec);
        ost.width(4);
        ost.fill('0');
        ost << listfile_ip;
        list_write_ascii(ost.str(), 5);
      }

      while ((ch = src->get())) {
        if (src->good()) {
          if (ch == '\n') {
            linenr++;
            // Automagically discard empty lines
            if (pos == linelength)
              goto listwrite;
            break;
          }
          line.append(1, static_cast<char>(ch));
          linelength++;
        } else {
          if (! src->eof()) {
            throw (Error::file_read_error(this->srcname));
          } else {
            if (linelength)
              break;
            return EOF;
          }
        }
      }
    }
    ch = line[pos++];
    return ch;
  }

  // void SourceHandler::list_write_hex(unsigned int, int n)
  // Writes the word value to the list file. It is a Hex representation of
  // the value that is left justified by n spaces.
  void SourceHandler::list_write_hex(unsigned int val, int n) 
  {
    std::ostringstream ost;

    ost.setf(std::ios::basefield, std::ios::hex);
    ost.unsetf(std::ios::dec);
    ost.width(4);
    ost.fill('0');
    ost << val;

    // fill it with spaces
    n -= 4;
    while (n-- > 0) {
      ost << ' ';
    }

    // put it in the list
    listfile.push_back(ost.str());
  }

  // void SourceHandler::list_write_ascii(const std::string, int n)
  // Writes the ascii string to the list file.
  void SourceHandler::list_write_ascii(const std::string &s, int n)
  {
    std::ostringstream ost;

    ost << s;

    // fill it with spaces
    n -= s.length();
    while (n-- > 0) {
      ost << ' ';
    }
    listfile.push_back(ost.str());
  }
  
  // void list_write() throw(Error::file_write_error)
  // Prints out the contents of the listfile.
  void SourceHandler::list_write() throw(Error::file_write_error)
  {
    if (::listfile) {
      for (std::list<std::string>::const_iterator it = listfile.begin();
           it != listfile.end(); it++) {
        *lst << *it;
        if (lst->fail())
          throw(Error::file_write_error(this->lstname));
      }
    }
  }

  // void SourceHandler::obj_write_hex(unsigned int, int n)
  // Writes the word value to the obj file. It is a Hex representation of
  // the value that is left justified by n spaces.
  void SourceHandler::obj_write_hex(unsigned int val, int n) 
  {
    std::ostringstream ost;

    ost.setf(std::ios::basefield, std::ios::hex);
    ost.unsetf(std::ios::dec);
    ost.width(4);
    ost.fill('0');
    ost << val;

    // fill it with spaces
    n -= 4;
    while (n-- > 0) {
      ost << ' ';
    }

    // put it in the obj
    objfile.push_back(ost.str());
  }

  // void SourceHandler::obj_write_ascii(const std::string &, int n)
  // Writes the ascii string to the obj file.
  void SourceHandler::obj_write_ascii(const std::string &s, int n)
  {
    std::ostringstream ost;

    ost << s;

    // fill it with spaces
    n -= s.length();
    while (n-- > 0) {
      ost << ' ';
    }
    objfile.push_back(ost.str());
  }
  
  // void obj_write() throw(Error::file_write_error)
  // Prints out the contents of the objfile.
  void SourceHandler::obj_write() throw(Error::file_write_error)
  {
    for (std::list<std::string>::const_iterator it = objfile.begin();
         it != objfile.end(); it++) {
      *obj_output << *it;
      if (obj_output->fail())
        throw(Error::file_write_error(this->objname));
    }
  }
}
