/* $Id: main.cpp,v 1.8 2004/11/21 18:52:51 markus Exp $ */

/*
 * Copyright (c) 2003 Markus Hennecke
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the Author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * The main program. Initializes the objects and starts the parser.
 */

// program includes:
#include "SymbolTable.hpp"
#include "SourceHandler.hpp"
#include "CodeGen.hpp"
#include "Parser.hpp"
#include "global_options.hpp"

// system includes:
#include <cstdlib>
#include <climits>
#include <string>
#include <iostream>

// global options:
bool        listfile = true;
bool        listfileallowed = true;
std::string objfile, srcfile, lstfile;
int         max_passes = 100;

// static void usage()
// Prints out how to use the program and exits with an exit code of 1.
static void usage()
{
  std::cout << "HASI - Hypothetical Machinecode Assembler and Simulator"
            << std::endl << std::endl
            << "as - HASI assembler" << std::endl;
  std::cout << "Copyright (c) 2003 Markus Hennecke" << std::endl
            << VERSION << std::endl << std::endl;
  
  std::cout << "USAGE: as [options] [-o outfile] file" << std::endl;
  std::cout << "\t-h           this help" << std::endl;
  std::cout << "\t-l           generate a listfile" << std::endl;
  std::cout << "\t-nl          generate no listfile" << std::endl;
  std::cout << "\t-m number    number of maximum passes (min 2 and max 100)" 
            << std::endl;
  std::cout << "\t-o filename  filename of the generated objectfile,"
            << " - for stdout" << std::endl;

  exit(1);
}

// static void parse_commandline(int argc, char **argv)
// Parses the commandline arguments. If something is not correct the usage
// information is printed and the program is terminated with an exit code 
// of 1.
static void parse_commandline(int argc, char **argv)
{
  for (int i = 1; i < argc; i++) {
    std::string opt(argv[i]);
    if (opt == "-h") {
      usage();
    } else if (opt == "-l") {
      listfile = true;
    } else if (opt == "-nl") {
      listfile = false;
    } else if (opt == "-m") {
      if (++i >= argc)
        usage();
      char *endptr = NULL;
      max_passes = strtol(argv[i], &endptr, 10);
      if (*endptr != '\0')
        usage();
      if (max_passes < 2) 
        usage();
    } else if (opt == "-o") {
      if (++i >= argc)
        usage();
      objfile = argv[i];
    } else {
      srcfile = argv[i++];
      if (! objfile.length()) {
        objfile = srcfile;
        std::string::size_type point = objfile.rfind('.');
        if (point != std::string::npos) {
          std::string::size_type l = objfile.length() - point;
          objfile.replace(objfile.length()-l, l, ".obj");
        } else {
          objfile.append(".obj");
        }
      }
      if (i != argc)
        usage();
    }
  }
  if (srcfile == "")
    usage();
  lstfile = srcfile;
  std::string::size_type point = lstfile.rfind('.');
  if (point != std::string::npos) {
    std::string::size_type l = lstfile.length() - point;
    lstfile.replace(lstfile.length()-l, l, ".lst");
  } else {
    lstfile.append(".lst");
  }
}

// int main(int, char **)
// The main program.
int main(int argc, char **argv)
{
  // check for correct arguments
  parse_commandline(argc, argv);

  try {
    Sourcehandler::SourceHandler SH(srcfile, lstfile, objfile);
    Lexer::LexicalAnalyzer LA(SH);
    SymbolTable::SymbolTable ST;
    CodeGenerator::CodeGen GEN(SH);

    Parser::Parser p(SH, LA, GEN, ST);

    int pass = 1;
    p.parse(pass++, false);
    p.rewind();
    while (p.parse(pass, (pass == max_passes))) {
      if (++pass > max_passes)
        break;
      p.rewind();
    }

    // write the output files
    SH.list_write();
    SH.obj_write();

    // Summary
    std::cerr << "Passes: " << pass << std::endl;

#ifdef DEBUG
    ST.dump();
#endif
  }
  catch (Error::error &e) {
    e.emit();
  }
  catch (...) {
    std::cerr << "Unexpected Error. Please contact the maintainer"
              << " of this software." << std::endl;
  }

  return 0;
}
