/* $Id: MainForm.cpp,v 1.28 2003/12/12 09:47:26 markus Exp $ */

/*
 * Copyright (c) 2003 Markus Hennecke
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the Author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,

 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

// program includes:
#include "MainForm.hpp"
#include "MemoryChain.hpp"
#include "CPU.hpp"
#include "ObjectFile.hpp"
#include "Options.hpp"
#include "QAssemblerOut.hpp"
#include "QAboutDialog.hpp"
#include "QHelpWindow.hpp"

// system includes:
#include <qlabel.h>
#include <qstatusbar.h>
#include <qlayout.h>
#include <qsplitter.h>
#include <qfiledialog.h>
#include <qmessagebox.h>
#include <qinputdialog.h>

#ifdef _WIN32
#ifdef _MSC_VER
#define popen(c,m)   _popen(c,m)
#define pclose(f)    _pclose(f)
#endif
#endif

namespace GUI {

  // MainForm::MainForm(Interpreter::CPU &cpu, QWidget *parent,
  //                    const char *name, WFlags fl)
  //   : QMainWindow(parent, name, fl), cpu(cpu)
  // The main windows. This widget creates all child widgets. Both the 
  // editor- and the debugger-page are created in the constructor. The
  // widgetstack is used to switch between them.
  MainForm::MainForm(Interpreter::CPU &cpu, QWidget *parent,
                     const char *name, WFlags fl)
    : QMainWindow(parent, name, fl), cpu(cpu), help(NULL), Debugger(false)
  {
    // create a status bar and give us a name
    statusBar();
    if (! name)
      setName("MainForm");

    setCentralWidget(new QWidget(this));
    mainFormLayout = new QVBoxLayout(centralWidget(), 1, 6);

    mainWidgetStack = new QWidgetStack(centralWidget());
    mainWidgetStack->setFrameShape(QWidgetStack::NoFrame);
    mainWidgetStack->setFrameShadow(QWidgetStack::Plain);

    editorPage = new QWidget(mainWidgetStack);
    editorPageLayout = new QVBoxLayout(editorPage, 2, 6);
    sourceTextEdit = new QTextEdit(editorPage);
    sourceTextEdit->setVScrollBarMode(QTextEdit::AlwaysOn);
    sourceTextEdit->setFont(QFont("fixed"));
    sourceTextEdit->setTextFormat(Qt::PlainText);
    sourceTextEdit->setWordWrap(QTextEdit::NoWrap);
    editorPageLayout->addWidget(sourceTextEdit);
    mainWidgetStack->addWidget(editorPage, 0);

    debuggerPage = new QSplitter(QSplitter::Vertical, mainWidgetStack);
    QSplitter *splitter2 = new QSplitter(QSplitter::Horizontal, debuggerPage);
    disassTable = new QDisassemblerPane(cpu, splitter2);
    debugTable = new QDebugPane(cpu, splitter2);
    QSplitter *splitter3 = new QSplitter(QSplitter::Horizontal, debuggerPage);
    memoryTable = new QMemoryPane(splitter3);
    memoryTable->setMemory(&(cpu.getMemoryChain()));
    stackTable = new QStackTable(NULL, &cpu, splitter3);
    stackTable->setMem(&(cpu.getMemoryChain()));

    mainWidgetStack->addWidget(debuggerPage, 1);

    mainFormLayout->addWidget(mainWidgetStack);

    fileNewAction = new QAction(this, "File New Action");
    fileOpenAction = new QAction(this, "File Open Action");
    fileSaveAction = new QAction(this, "File Save Action");
    fileSaveAsAction = new QAction(this, "File SaveAs Action");
    //filePrintAction = new QAction(this, "File Print Action");
    fileExitAction = new QAction(this, "File Exit Action");
    editUndoAction = new QAction(this, "Edit Undo Action");
    editRedoAction = new QAction(this, "Edit Redo Action");
    editCutAction = new QAction(this);
    editCopyAction = new QAction(this);
    editPasteAction = new QAction(this);
    editFindAction = new QAction(this);
    helpContentsAction = new QAction(this);
    helpIndexAction = new QAction(this);
    helpAboutAction = new QAction(this);

    assemblerSwitchDebugAction = new QAction(this,
                                "Assembler Switch Debugger Action");
    assemblerAssembleAction = new QAction(this, "Assembler Assemble Action");
    assemblerOptionsAction = new QAction(this, "Assembler Options Action");
    
    debugSwitchEditAction = new QAction(this, "Debug Switch Editor Action");
    debugStepAction = new QAction(this, "Debug Step Action");
    debugNextAction = new QAction(this, "Debug Next Action");
    debugRunAction = new QAction(this, "Debug Run Action");

    menuBar = new QMenuBar(this, "Menu Bar");

    fileMenu = new QPopupMenu(this, "File Menu");

    fileNewAction->addTo(fileMenu);
    fileOpenAction->addTo(fileMenu);
    fileSaveAction->addTo(fileMenu);
    fileSaveAsAction->addTo(fileMenu);
    fileMenu->insertSeparator();
    //filePrintAction->addTo(fileMenu);
    fileMenu->insertSeparator();
    fileExitAction->addTo(fileMenu);
    menuBar->insertItem("", fileMenu, 0);

    editMenu = new QPopupMenu(this, "Edit Menu");

    editUndoAction->addTo(editMenu);
    editRedoAction->addTo(editMenu);
    editMenu->insertSeparator();
    editCutAction->addTo(editMenu);
    editCopyAction->addTo(editMenu);
    editPasteAction->addTo(editMenu);
    editMenu->insertSeparator();
    editFindAction->addTo(editMenu);
    menuBar->insertItem("", editMenu, 1);

    assemblerMenu = new QPopupMenu(this, "Assembler Menu");
    assemblerSwitchDebugAction->addTo(assemblerMenu);
    assemblerMenu->insertSeparator();
    assemblerAssembleAction->addTo(assemblerMenu);
    assemblerOptionsAction->addTo(assemblerMenu);
    menuBar->insertItem("", assemblerMenu, 2);

    debugMenu = new QPopupMenu(this, "Debug Menu");
    debugSwitchEditAction->addTo(debugMenu);
    debugStepAction->addTo(debugMenu);
    debugNextAction->addTo(debugMenu);
    debugRunAction->addTo(debugMenu);
    menuBar->insertItem("", debugMenu, 3);

    helpMenu = new QPopupMenu(this, "Help Menu");

    helpContentsAction->addTo(helpMenu);
    helpIndexAction->addTo(helpMenu);
    helpMenu->insertSeparator();
    helpAboutAction->addTo(helpMenu);
    menuBar->insertSeparator();
    menuBar->insertItem("", helpMenu, 4);

    languageChange();

    connect(fileNewAction, SIGNAL(activated()), this, SLOT(fileNew()));
    connect(fileOpenAction, SIGNAL(activated()), this, SLOT(fileOpen()));
    connect(fileSaveAction, SIGNAL(activated()), this, SLOT(fileSave()));
    connect(fileSaveAsAction, SIGNAL(activated()), this, SLOT(fileSaveAs()));
    //connect(filePrintAction, SIGNAL(activated()), this, SLOT(filePrint()));
    connect(fileExitAction, SIGNAL(activated()), this, SLOT(fileExit()));
    connect(editUndoAction, SIGNAL(activated()), this, SLOT(editUndo()));
    connect(editRedoAction, SIGNAL(activated()), this, SLOT(editRedo()));
    connect(editCutAction, SIGNAL(activated()), this, SLOT(editCut()));
    connect(editCopyAction, SIGNAL(activated()), this, SLOT(editCopy()));
    connect(editPasteAction, SIGNAL(activated()), this, SLOT(editPaste()));
    connect(editFindAction, SIGNAL(activated()), this, SLOT(editFind()));
    connect(helpIndexAction, SIGNAL(activated()), this, SLOT(helpIndex()));
    connect(helpContentsAction, SIGNAL(activated()), this, 
            SLOT(helpContents()));
    connect(helpAboutAction, SIGNAL(activated()), this, SLOT(helpAbout()));

    connect(assemblerSwitchDebugAction, SIGNAL(activated()), 
            this, SLOT(assemblerSwitchDebug())); 
    connect(assemblerAssembleAction, SIGNAL(activated()), 
            this, SLOT(assemblerAssemble())); 
    connect(assemblerOptionsAction, SIGNAL(activated()), 
            this, SLOT(assemblerOptions()));

    connect(debugSwitchEditAction, SIGNAL(activated()),
            this, SLOT(debugSwitchEdit())); 
    connect(debugStepAction, SIGNAL(activated()), debugTable, SLOT(step()));
    connect(debugNextAction, SIGNAL(activated()), debugTable, SLOT(next()));
    connect(debugRunAction, SIGNAL(activated()), debugTable, SLOT(run()));

    connect(memoryTable, SIGNAL(memoryChanged(unsigned int)),
            disassTable, SLOT(addressChanged(unsigned int)));
    connect(memoryTable, SIGNAL(memoryChanged(unsigned int)),
            stackTable, SLOT(changedMem(unsigned int)));
    connect(debugTable, SIGNAL(IPChanged(unsigned int)),
            disassTable, SLOT(addressChanged(unsigned int)));
    connect(debugTable, SIGNAL(SPChanged(unsigned int)),
            stackTable, SLOT(changedSP(unsigned int)));
    connect(debugTable, SIGNAL(MemChanged(unsigned int)),
            disassTable, SLOT(addressChanged(unsigned int)));
    connect(debugTable, SIGNAL(MemChanged(unsigned int)),
            memoryTable, SLOT(memoryWritten(unsigned int)));
    connect(debugTable, SIGNAL(Running(bool)),
            this, SLOT(Run(bool)));

    // init the textedit widget...
    connect(sourceTextEdit, SIGNAL(cursorPositionChanged(int, int)),
            this, SLOT(cursorPositionChanged(int, int)));
    connect(sourceTextEdit, SIGNAL(textChanged()),
            this, SLOT(textChanged()));
    connect(sourceTextEdit, SIGNAL(copyAvailable(bool)),
            this, SLOT(copyStatus(bool)));
    connect(sourceTextEdit, SIGNAL(undoAvailable(bool)),
            this, SLOT(undoStatus(bool)));
    connect(sourceTextEdit, SIGNAL(redoAvailable(bool)),
            this, SLOT(redoStatus(bool)));
    editCutAction->setEnabled(false);
    editCopyAction->setEnabled(false);
    //editPasteAction->setEnabled(false);
    editUndoAction->setEnabled(false);
    editRedoAction->setEnabled(false);

    debugSwitchEdit();
    documentChanged = false;
    fileSaveAction->setEnabled(false);
    fileSaveAsAction->setEnabled(false);

    opts = new Options(this, NULL);

    cursorPositionLabel = new QLabel(this);
    statusBar()->addWidget(cursorPositionLabel, 0, true);
    cursorPositionChanged(0, 0);
  }

  // MainForm::~MainForm()
  // The destructor... delete the options dialog if we created one.
  MainForm::~MainForm()
  {
    if (opts)
      delete opts;
  }

  // void MainForm::languageChange()
  // If we get multilingual this is the key.
  void MainForm::languageChange()
  {
    setCaption(tr("HASI - Hypothetical Machinecode Assembler and Simulator"));
    fileNewAction->setText(tr("New"));
    fileNewAction->setMenuText(tr("&New"));
    fileNewAction->setAccel(tr("Ctrl+N"));
    fileOpenAction->setText(tr("Open"));
    fileOpenAction->setMenuText(tr("&Open..."));
    fileOpenAction->setAccel(tr("Ctrl+O"));
    fileSaveAction->setText(tr("Save"));
    fileSaveAction->setMenuText(tr("&Save"));
    fileSaveAction->setAccel(tr("Ctrl+S"));
    fileSaveAsAction->setText(tr("Save As"));
    fileSaveAsAction->setMenuText(tr("Save &As..."));
    fileSaveAsAction->setAccel(QString::null);
    //filePrintAction->setText(tr("Print"));
    //filePrintAction->setMenuText(tr("&Print..."));
    //filePrintAction->setAccel(tr("Ctrl+P"));
    fileExitAction->setText(tr("Exit"));
    fileExitAction->setMenuText(tr("E&xit"));
    fileExitAction->setAccel(QString::null);
    editUndoAction->setText(tr("Undo"));
    editUndoAction->setMenuText(tr("&Undo"));
    editUndoAction->setAccel(tr("Ctrl+Z"));
    editRedoAction->setText(tr("Redo"));
    editRedoAction->setMenuText(tr("&Redo"));
    editRedoAction->setAccel(tr("Ctrl+Y"));
    editCutAction->setText(tr("Cut"));
    editCutAction->setMenuText(tr("&Cut"));
    editCutAction->setAccel(tr("Ctrl+X"));
    editCopyAction->setText(tr("Copy"));
    editCopyAction->setMenuText(tr("C&opy"));
    editCopyAction->setAccel(tr("Ctrl+C"));
    editPasteAction->setText(tr("Paste"));
    editPasteAction->setMenuText(tr("&Paste"));
    editPasteAction->setAccel(tr("Ctrl+V"));
    editFindAction->setText(tr("Find"));
    editFindAction->setMenuText(tr("&Find..."));
    editFindAction->setAccel(tr("Ctrl+F"));
    helpContentsAction->setText(tr("Contents"));
    helpContentsAction->setMenuText(tr("&Contents..."));
    helpContentsAction->setAccel(QString::null);
    helpIndexAction->setText(tr("Index"));
    helpIndexAction->setMenuText(tr("&Index..."));
    helpIndexAction->setAccel(QString::null);
    helpAboutAction->setText(tr("About"));
    helpAboutAction->setMenuText(tr("&About"));
    helpAboutAction->setAccel(QString::null);
    assemblerSwitchDebugAction->setText(tr("Switch to Debugger"));
    assemblerSwitchDebugAction->setMenuText(tr("Switch to De&bugger"));
    assemblerSwitchDebugAction->setAccel(QKeySequence(Qt::Key_F3));
    assemblerAssembleAction->setText(tr("Assemble"));
    assemblerAssembleAction->setMenuText(tr("&Assemble"));
    assemblerAssembleAction->setAccel(QKeySequence(Qt::Key_F4));
    assemblerOptionsAction->setText(tr("Options"));
    assemblerOptionsAction->setMenuText(tr("&Options..."));
    assemblerOptionsAction->setAccel(QString::null);
    debugSwitchEditAction->setText(tr("Switch to Editor"));
    debugSwitchEditAction->setMenuText(tr("Switch to &Editor"));
    debugSwitchEditAction->setAccel(QKeySequence(Qt::Key_F2));
    debugStepAction->setText(tr("Step"));    
    debugStepAction->setMenuText(tr("&Step"));
    debugStepAction->setAccel(QKeySequence(Qt::Key_F7));
    debugNextAction->setText(tr("Next"));    
    debugNextAction->setMenuText(tr("&Next"));
    debugNextAction->setAccel(QKeySequence(Qt::Key_F7 + SHIFT));
    debugRunAction->setText(tr("Run"));    
    debugRunAction->setMenuText(tr("&Run"));
    debugRunAction->setAccel(QKeySequence(Qt::Key_F8));

    menuBar->findItem(0)->setText(tr("&File"));
    menuBar->findItem(1)->setText(tr("&Edit"));
    menuBar->findItem(2)->setText(tr("&Assembler"));
    menuBar->findItem(3)->setText(tr("&Debug"));
    menuBar->findItem(4)->setText(tr("&Help"));
  }

  // void MainForm::fileNew()
  // If the document has changed the user is asked to erase the doc.
  void MainForm::fileNew()
  {
    bool newdoc = true;
    if (documentChanged) {
      if (QMessageBox::warning(this, tr("Document not saved"),
                               tr("The document has not been saved.\n"
                                  "Do you want to continue?"), 
                               QMessageBox::Yes, QMessageBox::No)
          == QMessageBox::No)
        newdoc = false;
    }
    if (newdoc) {
      sourceTextEdit->setText(QString::null);
      documentChanged = false;
      fileSaveAction->setEnabled(false);
      fileSaveAsAction->setEnabled(false);
      documentName = "";
      assemblerAssembleAction->setEnabled(false);
    }
  }

  // void MainForm::fileOpen()
  // If the debugger is active an object file can be opened. If the editor
  // is active source files can be loaded.
  void MainForm::fileOpen()
  {
    try {
      QFileDialog fileDialog(this);

      if (Debugger) {
    
        fileDialog.setFilter("*.obj");
        fileDialog.exec();
        if (QDialog::Accepted == fileDialog.result()) {
          std::string filename = fileDialog.selectedFile().ascii();

          Interpreter::ObjectFile obj(filename);
          obj.read(&(cpu.getMemoryChain()));
          disassTable->disassemble();
          // TODO: make the memory output redraw itself
        }
      } else {
        bool open = true;
        // first check if we really want this
        if (documentChanged) {
          if (QMessageBox::warning(this, tr("Document not saved"),
                                   tr("The document has not been saved.\n"
                                      "Do you want to continue?"), 
                                   QMessageBox::Yes, QMessageBox::No)
              == QMessageBox::No)
            open = false;
        }     
        
        // still wanting to open a new file?
        if (open) {
          fileDialog.setFilter("*.asm");
          fileDialog.exec();
          if (QDialog::Accepted == fileDialog.result()) {
            std::string filename = fileDialog.selectedFile().ascii();
            
            documentName = filename;
            assemblerAssembleAction->setEnabled(true);
            
            std::ifstream sourcefile(filename.c_str());
            std::string source;
            while (sourcefile) {
              std::string line;
              std::getline(sourcefile, line);
              source += line + "\n";
            }
            if (! sourcefile.eof())
              throw Error::file_read_error(filename);

            sourceTextEdit->setText(source.c_str());
            documentChanged = false;
            fileSaveAction->setEnabled(false);
            fileSaveAsAction->setEnabled(false);
          }
        }
      }
    }  
    catch (Error::file_open_error &e) {
      QString message;
      message.sprintf("Unable to open \"%s\".", e.getFilename().c_str());
      QMessageBox::critical(this, tr("Error!"), tr(message),
                            QMessageBox::Ok, QMessageBox::NoButton);
      documentName = "";
      assemblerAssembleAction->setEnabled(false);
      sourceTextEdit->setText(QString::null);
    }
    catch (Error::file_read_error &e) {
      QString message;
      message.sprintf("Read failed on \"%s\".", e.getFilename().c_str());
      QMessageBox::critical(this, tr("Error!"), tr(message),
                            QMessageBox::Ok, QMessageBox::NoButton);
    }
  }

  // void MainForm::fileSave()
  // Slot, saves the document...
  void MainForm::fileSave()
  {
    if (documentName.empty()) {
      fileSaveAs();
    } else {
      QFile file(documentName.c_str());
      if (file.open(IO_WriteOnly)) {
        QTextStream stream(&file);
        stream << sourceTextEdit->text();
        file.close();
        documentChanged = false;
        fileSaveAction->setEnabled(false);
        fileSaveAsAction->setEnabled(false);
	assemblerAssembleAction->setEnabled(true);
      }
    }
  }

  // void MainForm::textChanged()
  // Slot, gets called if the document changes so the Save menus can
  // be enabled.
  void MainForm::textChanged()
  {
    if (! documentChanged) {
      documentChanged = true;
      fileSaveAction->setEnabled(true);
      fileSaveAsAction->setEnabled(true);
    }
    // TODO: 
  }

  // void MainForm::cursorPositionChanged(int, int)
  // Slot, gets called if the cursor position changed with the mouse.
  void MainForm::cursorPositionChanged(int para, int pos)
  {
    // pos references the position in the string. The problem are tab stops
    QString s(sourceTextEdit->text());

    for (int i = 0; i < para; i++) {
      int newline = s.find(QString("\n"), 0);
      if (newline != -1) {
	s.replace(0, newline, "");
      }
    }

    int count = 0;
    for (int i = 0; i < pos; i++) {
      count += (s[i] == '\t') ? 8 : 1;
    }

    // para references the line beginning with 0
    QString cur_pos = "Pos: " + QString::number(++para) + "," +
      QString::number(count);
    cursorPositionLabel->setText(cur_pos);
  }
    

  // void MainForm::fileSaveAs()
  // Asks the user for a new filename before saving...
  void MainForm::fileSaveAs()
  {
    QString s = QFileDialog::getSaveFileName(
                    ".",
                    "Sourcefiles (*.asm)",
                    this,
                    "save file dialog"
                    "Choose a filename to save under" );
    if (! s.isEmpty()) {
      documentName = s.ascii();
      QFile file(documentName.c_str());
      if (file.open(IO_WriteOnly)) {
        QTextStream stream(&file);
        stream << sourceTextEdit->text();
        file.close();
        documentChanged = false;
        fileSaveAction->setEnabled(false);
        fileSaveAsAction->setEnabled(false);
        assemblerAssembleAction->setEnabled(true);
      }
    }
  }

#if 0
  // void MainForm::filePrint()
  // Prints the file, well should print the file is better to say...
  void MainForm::filePrint()
  {
    qWarning("MainForm::filePrint(): Not implemented yet\n"
             "No way to implement this without a printer :-(");
    QPrinter prt(QPrinter::PrinterResolution);
    prt.setDocname(documentName);
  }
#endif

  // void MainForm::fileExit()
  // Asks the user if the document has changed if it should really quit.
  void MainForm::fileExit()
  {
    this->close();
  }

  // void MainForm::closeEvent(QCloseEvent *)
  // Gets called if the window is closed.
  void MainForm::closeEvent(QCloseEvent *e)
  {
    bool quit = true;
    if (documentChanged) {
      if (QMessageBox::warning(this, tr("Document not saved"),
                               tr("The document has not been saved.\n"
                                  "Do you want to quit?"), 
                               QMessageBox::Yes, QMessageBox::No)
          == QMessageBox::No)
        quit = false;
    }
    if (quit) {
      opts->saveOptions();
      if (help)
        help->close();
      e->accept();
    }
  }

  // void MainForm::editUndo()
  void MainForm::editUndo()
  {
    sourceTextEdit->undo();
  }

  // void MainForm::editRedo()
  void MainForm::editRedo()
  {
    sourceTextEdit->redo();
  }

  // void MainForm::editCut()
  void MainForm::editCut()
  {
    sourceTextEdit->cut();
    editPasteAction->setEnabled(true);
  }

  // void MainForm::editCopy()
  void MainForm::editCopy()
  {
    sourceTextEdit->copy();
    editPasteAction->setEnabled(true);
  }

  // void MainForm::editPaste()
  void MainForm::editPaste()
  {
    sourceTextEdit->paste();
  }

  // void MainForm::editFind()
  void MainForm::editFind()
  {
    bool ok;
    QString s = QInputDialog::getText(tr("Find"),
                                      tr("String to find:"),
                                      QLineEdit::Normal,
                                      QString::null, &ok, this);
    if (ok && !s.isEmpty()) {
      if (sourceTextEdit->find(s, true, true))
        // TODO: search again etc...
        ;
    }
  }

  // void MainForm::helpIndex()
  // Shows the helpindex in the help window.
  void MainForm::helpIndex()
  {
    if (! help) {
      help = new QHelpWindow("index.html", opts->getHelpPath().c_str(), this);
    } else {
      help->changeSource("index.html");
    }
    help->show();
  }

  // void MainForm::helpContents()
  // Shows the help windows with the starting documentation.
  void MainForm::helpContents()
  {
    if (! help) {
      help = new QHelpWindow("user-guide-pure.html", 
			     opts->getHelpPath().c_str(), this);
    } else {
      help->changeSource("user-guide-pure.html");
    }
    help->show();
  }

  // void MainForm::helpAbout()
  void MainForm::helpAbout()
  {
    QAboutDialog about(this);
    about.exec();
  }

  // void MainForm::assemblerSwitchDebug()
  // Slot, called to switch to the debugging context.
  void MainForm::assemblerSwitchDebug()
  {
    Debugger = true;
    menuBar->setItemEnabled(1, false);
    menuBar->setItemEnabled(2, false);
    menuBar->setItemEnabled(3, true);
    fileNewAction->setEnabled(false);
    fileSaveAction->setEnabled(false);
    fileSaveAsAction->setEnabled(false);
    //filePrintAction->setEnabled(false);
    mainWidgetStack->raiseWidget(debuggerPage);
  }

  // void MainForm::assemblerSwitchEdit()
  // Slot, called to switch to the editor.
  void MainForm::debugSwitchEdit()
  {
    Debugger = false;
    menuBar->setItemEnabled(1, true);
    menuBar->setItemEnabled(2, true);
    menuBar->setItemEnabled(3, false);
    fileNewAction->setEnabled(true);
    if (documentChanged) {
      fileSaveAction->setEnabled(true);
      fileSaveAsAction->setEnabled(true);
    } else {
      fileSaveAction->setEnabled(false);
      fileSaveAsAction->setEnabled(false);
    }      
    //filePrintAction->setEnabled(true);
    mainWidgetStack->raiseWidget(editorPage);
  }

  // void MainForm::assemblerAssemble()
  // Slot, calls the assembler and redirects output to a new window.
  void MainForm::assemblerAssemble()
  {
    // nothing to do if we have no document
    if (documentName.empty()) {
      QMessageBox::information(this, tr("No sourcecode available"),
                               tr("There is no sourcecode available, aborting"),
                               QMessageBox::Ok, QMessageBox::NoButton);
      return;
    }
    // save the file if not already done
    if (documentChanged) {
      fileSave();
    }

    // build the command string
    std::string cmd = "'" + opts->getAsPath();
#ifdef _WIN32
    cmd += "\\as.exe' ";
#else
    cmd += "/as' ";
#endif
    if (opts->getListfile()) {
      cmd += "-l ";
    } else {
      cmd += "-nl ";
    }
    std::string objfile = documentName;
    std::string::size_type point = objfile.rfind('.');
    if (point != std::string::npos) {
      std::string::size_type l = objfile.length() - point;
      objfile.replace(objfile.length()-l, l, ".obj");
    } else {
      objfile.append(".obj");
    }
    cmd += " -o '";
    cmd += objfile;
    cmd += "' -m ";

    // QString is an easy way to format numbers in a save way...
    QString passesString;
    passesString.setNum(opts->getPasses(), 10);
    cmd += passesString.ascii();
    
    cmd += " '";
    cmd += documentName;
    cmd += "' 2>&1";

    // Redirect the output to a file handle so we can capture it
    FILE *fd = popen(cmd.c_str(), "r");

    // Create the widget that should hold the output
    QAssemblerOutput *output = new QAssemblerOutput(this);

    char c;
    std::string line = "";
    while ((c = fgetc(fd)) != EOF) {
      line += c;
      if (c == '\n') {
        output->addLine(line);
        line = "";
      }
    }
    // close the filehandle
    fclose(fd);
    
    if (output->exec() == QDialog::Accepted) {
      // Go to the debugger and load the object file if the button
      // "Go to debugger" was pressed.
      Interpreter::ObjectFile obj(objfile);
      obj.read(&(cpu.getMemoryChain()));
      disassTable->disassemble();

      delete output;
      assemblerSwitchDebug();
    } else {
      delete output;
    }
  }

  // void MainForm::assemblerOptions()
  // Shows the dialog to alter the options of the assember.
  void MainForm::assemblerOptions()
  {
    opts->exec();
  }

  // void MainForm::Run(bool running)
  // Slot, gets connected to receive a signal that the disassembler and
  // memory tables are to be disabled or enabled.
  void MainForm::Run(bool running)
  {
    disassTable->setEnabled(running);
    memoryTable->setEnabled(running);
  }

  // void copyStatus(bool available)
  // Slot, get called if the copy status changes.
  void MainForm::copyStatus(bool available)
  {
    if (available) {
      editCopyAction->setEnabled(true);
      editCutAction->setEnabled(true);
    } else {
      editCopyAction->setEnabled(false);
      editCutAction->setEnabled(false);
    }
  }

  // void MainForm::redoStatus(bool available)
  // Slot, gets called if undo is available at the textedit.
  void MainForm::undoStatus(bool available)
  {
    if (available) {
      editUndoAction->setEnabled(true);
    } else {
      editUndoAction->setEnabled(false);
    }
  }

  // void MainForm::redoStatus(bool available)
  // Slot, gets called if redo is available at the textedit.
  void MainForm::redoStatus(bool available)
  {
    if (available) {
      editRedoAction->setEnabled(true);
    } else {
      editRedoAction->setEnabled(false);
    }
  }
}
