/* $Id: QDebugPane.cpp,v 1.12 2003/12/12 09:47:26 markus Exp $ */

/*
 * Copyright (c) 2003 Markus Hennecke
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the Author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * The declaration of a Widget that contains the actual debugger. The CPU
 * information and the breakpoint list.
 */

#include "Error.hpp" // workaround

// system includes:
#include <qpushbutton.h>
#include <qlayout.h>
#include <qtooltip.h>
#include <qgroupbox.h>
#include <qvgroupbox.h>
#include <qlabel.h>
#include <qlistview.h>
#include <qmessagebox.h>

#include <string>
#include <list>

// program includes:
#include "QDebugPane.hpp"
#include "QDialogAddress.hpp"

namespace GUI {

  // class Breakpoint
  // Declared here because it is local to the Debug Panel class.
  class Breakpoint
  {
  private:
    // This will hold a address and a comment.
    unsigned int address;
    std::string comment;
  public:
    Breakpoint(const unsigned int address, const std::string &comment)
      : address(address), comment(comment)
    { }
    
    unsigned int getAddress() const { return address; }
    const std::string &getComment() const { return comment; }
  };

  static std::list<Breakpoint> bpList;

  // static bool is_Breakpoint(const unsigned int address)
  // Returns true if the address is associated with a breakpoint.
  static bool is_Breakpoint(const unsigned int address)
  {
    std::list<Breakpoint>::const_iterator it = bpList.begin();
    while (it != bpList.end()) {
      if (it->getAddress() == address)
        return true;
      it++;
    }
    return false;
  }

  // QDebugPane::QDebugPane(QWidget *parent, const char *name)
  //   : QWidget(parent, name)
  // Constructor that constructs the widgets child widgets and sets everything
  // up.
  QDebugPane::QDebugPane(Interpreter::CPU &cpu, 
                         QWidget *parent, const char *name)
    : QWidget(parent, name), cpu(cpu), stop(false)
  {
    // We have a vertical main Layout.
    QVBoxLayout *mainLayout = new QVBoxLayout(this);
    QHBoxLayout *upperLayout = new QHBoxLayout(mainLayout);

    QGroupBox *cpuBox = new QGroupBox(tr("CPU Register"), this);
    cpuBox->setMinimumWidth(250);
    
    // TODO: next line's just a hack to make it look better...
    // this should be fixed in a better way.
    QVBoxLayout *tmp = new QVBoxLayout(cpuBox, 22, 5);
    QGridLayout *cpuBoxLineLayout = new QGridLayout(5, 2, 0, "CPU Registers");
    tmp->addLayout(cpuBoxLineLayout);
    QLabel *l = new QLabel(tr("Akku:"), cpuBox);
    cpuAkku = new QHexLineEdit(cpuBox);
    cpuBoxLineLayout->addWidget(l, 0, 0);
    cpuBoxLineLayout->addWidget(cpuAkku, 0, 1);
    QToolTip::add(l,
                  tr("Akkumulator Register"));
    
    l = new QLabel(tr("Help:"), cpuBox);
    cpuSwap = new QHexLineEdit(cpuBox);
    cpuBoxLineLayout->addWidget(l, 1, 0);
    cpuBoxLineLayout->addWidget(cpuSwap, 1, 1);
    QToolTip::add(l, tr("Swap Register"));

    l = new QLabel(tr("Index:"), cpuBox);
    cpuIndex = new QHexLineEdit(cpuBox);
    cpuBoxLineLayout->addWidget(l, 2, 0);
    cpuBoxLineLayout->addWidget(cpuIndex, 2, 1);
    QToolTip::add(l, tr("Index Register"));

    l = new QLabel(tr("SP:"), cpuBox);
    cpuSP = new QHexLineEdit(cpuBox);
    cpuBoxLineLayout->addWidget(l, 3, 0);
    cpuBoxLineLayout->addWidget(cpuSP, 3, 1);
    QToolTip::add(l, tr("Stack Pointer"));

    l = new QLabel(tr("IP:"), cpuBox);
    cpuIP = new QHexLineEdit(cpuBox);
    cpuBoxLineLayout->addWidget(l, 4, 0);
    cpuBoxLineLayout->addWidget(cpuIP, 4, 1);
    QToolTip::add(l, tr("Instruction Pointer"));

    tmp->addStretch();

    upperLayout->addWidget(cpuBox);

    QVGroupBox *debugBox = new QVGroupBox(tr("Debugger"), this);
    stepButton = new QPushButton(tr("Step"), debugBox);
    QToolTip::add(stepButton,
                  tr("Step one instruction"));
    nextButton = new QPushButton(tr("Next"), debugBox);
    QToolTip::add(nextButton,
                  tr("Stop over next instruction"));
    runButton = new QPushButton(tr("Run"), debugBox);
    QToolTip::add(runButton,
                  tr("Run from IP"));
    stopButton = new QPushButton(tr("Stop"), debugBox);
    stopButton->setEnabled(false);
    QToolTip::add(stopButton,
                  tr("Stop Execution"));
    newBreakpoint = new QPushButton(tr("Add Breakpoint"), debugBox);
    QToolTip::add(newBreakpoint,
                  tr("Add Breakpoint"));
    deleteBreakpoint = new QPushButton(tr("Delete Breakpoint"), debugBox);
    deleteBreakpoint->setEnabled(false);
    QToolTip::add(deleteBreakpoint,
                  tr("Delete selected breakpoint"));
    upperLayout->addWidget(debugBox);
    
    breakpointList = new QListView(this);
    breakpointList->addColumn(tr("Address"));
    breakpointList->addColumn(tr("Comment"));

    QToolTip::add(breakpointList,
                  tr("Breakpoint List"));
    mainLayout->addWidget(breakpointList);

    updateDisplay();

    connect(cpuAkku, SIGNAL(returnPressed()), this, SLOT(changedAkku()));
    connect(cpuSwap, SIGNAL(returnPressed()), this, SLOT(changedHelp()));
    connect(cpuIndex, SIGNAL(returnPressed()), this, SLOT(changedIndex()));
    connect(cpuSP, SIGNAL(returnPressed()), this, SLOT(changedSP()));
    connect(cpuIP, SIGNAL(returnPressed()), this, SLOT(changedIP()));

    connect(stepButton, SIGNAL(clicked()),
            this, SLOT(step()));
    connect(nextButton, SIGNAL(clicked()),
            this, SLOT(next()));
    connect(runButton, SIGNAL(clicked()),
            this, SLOT(run()));
    connect(stopButton, SIGNAL(clicked()),
            this, SLOT(stopPressed()));
    connect(newBreakpoint, SIGNAL(clicked()),
            this, SLOT(addBreakpoint()));
    connect(deleteBreakpoint, SIGNAL(clicked()),
            this, SLOT(remBreakpoint()));
    connect(breakpointList, SIGNAL(selectionChanged()),
            this, SLOT(breakpointSelected()));
  }

  // QDebugPane::~QDebugPane()
  // Nothing to do here.
  QDebugPane::~QDebugPane()
  { }

  // void QDebugPane::updateDisplay()
  // Slot, is addressed if the cpu has completed a step.
  void QDebugPane::updateDisplay()
  {
    cpuAkku->setNumber(cpu.getAkku());
    cpuIndex->setNumber(cpu.getIndex());
    cpuSP->setNumber(cpu.getStackPointer());
    cpuSwap->setNumber(cpu.getHelp());
    cpuIP->setNumber(cpu.getInstructionPointer());
  }

  // void QDebugPane::changedAkku()
  // Slot, is addressed if the lineedit for the cpus akkumulator
  // has new input (finished with a return)
  void QDebugPane::changedAkku()
  {
    int newval = cpuAkku->getNumber();
    cpu.setAkku(static_cast<short>(newval));
  }

  // void QDebugPane::changedHelp()
  // Slot, is addressed if the lineedit for the cpus help register
  // has new input (finished with a return)
  void QDebugPane::changedHelp()
  {
    int newval = cpuSwap->getNumber();
    cpu.setHelp(static_cast<short>(newval));
  }

  // void QDebugPane::changedIndex()
  // Slot, is addressed if the lineedit for the cpus index register
  // has new input (finished with a return)
  void QDebugPane::changedIndex()
  {
    int newval = cpuIndex->getNumber();
    cpu.setIndex(static_cast<short>(newval));
  }

  // void QDebugPane::changedSP()
  // Slot, is addressed if the lineedit for the cpus stack pointer
  // has new input (finished with a return)
  void QDebugPane::changedSP()
  {
    int newval = cpuSP->getNumber();
    cpu.setStackPointer(static_cast<unsigned int>(newval));
    emit(SPChanged(static_cast<unsigned int>(newval)));
  }

  // void QDebugPane::changedIP()
  // Slot, is addressed if the lineedit for the cpu instruction pointer
  // has new input (finished with a return)
  void QDebugPane::changedIP()
  {
    int newval = cpuIP->getNumber();
    cpu.setInstructionPointer(static_cast<unsigned int>(newval));
    emit(IPChanged(static_cast<unsigned int>(newval)));
  }

  // void QDebugPane::step()
  // Steps the cpu one step.
  void QDebugPane::step()
  {
    try {
      cpu.halt(false);
      cpu.step();
      emit(IPChanged(cpu.getInstructionPointer()));
      emit(SPChanged(cpu.getStackPointer()));
      emit(MemChanged(cpu.getLastWriteAddress()));
      updateDisplay();
    }
    catch (Error::division_by_zero &e) {
      QMessageBox::critical(this, tr("Division by Zero"), 
                            tr("Division by Zero, program execution haltet."),
                            QMessageBox::Ok, QMessageBox::NoButton);
    }
    catch (...) {
      QMessageBox::critical(this, tr("Internal Error"), 
                            tr("Please submit a full bugreport\nto the maintainer of this Software"),
                            QMessageBox::Ok, QMessageBox::NoButton);
    }
  }

  // void QDebugPane::next()
  // Runs the cpu from the current Instruction Pointer. If it is a normal
  // instruction this function behaves like step. If it is a subroutine
  // that is called the execution will be continued until either the cpu
  // is halted, the instruction after the starting instruction or a 
  // breakpoint is reached.
  void QDebugPane::next()
  {
    prepareRun(false);
    try {
      cpu.halt(false);
      stop = false;

      // hm, no way to determine the address of the next instruction
      // but it is only 1 or 2 bytes long
      unsigned int retAddress = cpu.getInstructionPointer() + 1;

      while (!stop) {
        cpu.step();

        if ((cpu.getInstructionPointer() == retAddress) ||
            (cpu.getInstructionPointer() == (retAddress + 1))) {
          break;
        }

        // Stop if the cpu is halted
        if (cpu.isHalted())
          break;

        // Stop if the next IP is a Breakpoint
        if (is_Breakpoint(cpu.getInstructionPointer()))
          break;
      }
    }
    catch (...) {
      QMessageBox::critical(this, tr("Internal Error"), 
                            tr("Please submit a full bugreport\nto the maintainer of this Software"),
                            QMessageBox::Ok, QMessageBox::NoButton);
    }

    emit(IPChanged(cpu.getInstructionPointer()));
    emit(SPChanged(cpu.getStackPointer()));
    emit(MemChanged(cpu.getLastWriteAddress()));
    updateDisplay();

    prepareRun(true);
    stop = false;
  }

  // void QDebugPane::prepareRun(bool)
  void QDebugPane::prepareRun(bool en)
  {
    if (breakpointList->selectedItem()) {
      if (en) {
        deleteBreakpoint->setEnabled(true);
      } else {
        deleteBreakpoint->setEnabled(false);
      }
    } else {
      deleteBreakpoint->setEnabled(false);
    }
    newBreakpoint->setEnabled(en);
    breakpointList->setEnabled(en);
    stepButton->setEnabled(en);
    nextButton->setEnabled(en);
    runButton->setEnabled(en);
    stopButton->setEnabled(! en);
    cpuAkku->setEnabled(en);
    cpuSwap->setEnabled(en);
    cpuIndex->setEnabled(en);
    cpuSP->setEnabled(en);
    cpuIP->setEnabled(en);

    emit(Running(en));
  }

  // void QDebugPane::run()
  // Runs the cpu until either the cpu is halted or a breakpoint is
  // reached.
  void QDebugPane::run()
  {
    prepareRun(false);
    try {
      cpu.halt(false);
      stop = false;
      
      while ((! cpu.isHalted()) && (! stop)) {
        cpu.step();

        // Stop if the next IP is a Breakpoint
        if (is_Breakpoint(cpu.getInstructionPointer()))
          break;
      }
    }
    catch (...) {
      QMessageBox::critical(this, tr("Internal Error"), 
                            tr("Please submit a full bugreport\nto the maintainer of this Software"),
                            QMessageBox::Ok, QMessageBox::NoButton);
    }
    emit(IPChanged(cpu.getInstructionPointer()));
    emit(SPChanged(cpu.getStackPointer()));
    emit(MemChanged(cpu.getLastWriteAddress()));
    updateDisplay();

    prepareRun(true);
    stop = false;
  }

  // void addBreakpoint()
  // Opens a Dialog that asks the user for an address and a comment.
  // This information is added to the breakpoint list.
  void QDebugPane::addBreakpoint()
  {
    QDialogAddress dlg(cpu.getInstructionPointer(), "", this);
    if (dlg.exec() == QDialog::Accepted) {
      // Check if there is already a breakpoint with this address
      if (! is_Breakpoint(dlg.getAddress())) {
        Breakpoint bp(dlg.getAddress(), dlg.getComment());
        bpList.push_front(bp);
        QString s;
        s.sprintf("%04x", dlg.getAddress());
        new QListViewItem(breakpointList, s, dlg.getComment().c_str());
      } else {
        QMessageBox::information(this, tr("Breakpoint already defined"),
                                 tr("Breakpoint already defined.\n Not added to Breakpoints"),
                                 QMessageBox::Ok);
      }
    }
  }

  // void QDebugPane::remBreakpoint()
  // Removes a breakpoint from the listview and the internal breakpoint
  // list.
  void QDebugPane::remBreakpoint()
  {
    QListViewItem *selected = breakpointList->selectedItem();
    if (selected) {
      bool ok;
      QString addr = selected->text(0);
      unsigned int address = addr.toUInt(&ok, 16);
      for (std::list<Breakpoint>::iterator it = bpList.begin();
           it != bpList.end(); it++) {
        if (it->getAddress() == address) {
          bpList.erase(it);
          break;
        }
      }
      breakpointList->takeItem(selected);
    }
  }

  // void QDebugPane::breakpointSelected()
  // Slot, called if the selection in the breakpoint list view changed.
  void QDebugPane::breakpointSelected()
  {
    QListViewItem *selected = breakpointList->selectedItem();
    if (selected) {
      deleteBreakpoint->setEnabled(true);
    } else {
      deleteBreakpoint->setEnabled(false);
    }
  }

  // void QDebugPane::stopPressed()
  // Sets the flag that the program has to be stopped after the next step.
  void QDebugPane::stopPressed()
  {
    stop = true;
    stopButton->setEnabled(false);
  }
}
