/* $Id: QDisassemblerPane.cpp,v 1.7 2003/12/12 09:47:27 markus Exp $ */

/*
 * Copyright (c) 2003 Markus Hennecke
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the Author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * The implementation of a Widget that contains a QTable with disassembled
 * memory.
 */

// Weired bug, did not find out why Error.hpp must be included before any
// QT includes.
#include "Error.hpp"

// system includes:
#include <qlayout.h>
#include <qtable.h>
#include <qtooltip.h>

#include <string>

// program includes:
#include "QDisassemblerPane.hpp"

#define MAXLINES 50

namespace GUI {

  class QLeftJustifiedTableItem : public QTableItem
  {
  public:
    QLeftJustifiedTableItem(QTable *table, EditType et, const QString &text)
      : QTableItem(table, et, text)
    { }
    virtual int alignment() const { return Qt::AlignLeft; }
  };

  // QDisassemblerPane::QDisassemblerPane(Interpreter::CPU &cpu, 
  //                                      QWidget *parent, const char *name)
  //   : QWidget(parent, name)
  // Constructor that constructs the widgets child widgets and sets everything
  // up.
  QDisassemblerPane::QDisassemblerPane(Interpreter::CPU &cpu, QWidget *parent,
                                       const char *name)
    : QWidget(parent, name), cpu(cpu)
  {
    // We have a vertical main Layout.
    setMinimumWidth(300);
    setMinimumHeight(200);
    QVBoxLayout *mainLayout = new QVBoxLayout(this);

    disassemblerTable = new QTable(50, 3, this);
    mainLayout->addWidget(disassemblerTable);
    QToolTip::add(disassemblerTable,
                  tr("Disassembled Memory"));

    // Create the labels for the columns
    QStringList colLabels;
    colLabels.push_back(tr("Value"));
    colLabels.push_back(tr("Opcode"));
    colLabels.push_back(tr("Argument"));
    disassemblerTable->setColumnLabels(colLabels);
    
    // Create all Items that are in the table
    for (int i = 0; i < MAXLINES; i++) {
      disassemblerTable->setRowStretchable(i, false);
      for (int j = 0; j < 3; j++) {
        // Add our own TableItems to have a left justified output in the
        // whole table.
        QLeftJustifiedTableItem *ti =
          new QLeftJustifiedTableItem(disassemblerTable, QTableItem::Never,
                                      "");
        disassemblerTable->setItem(i, j, ti);
      }
    }
    // Make it look nice :-)
    disassemblerTable->setReadOnly(true);
    disassemblerTable->setSelectionMode(QTable::SingleRow);
    disassemblerTable->setFocusStyle(QTable::FollowStyle);
    disassemblerTable->setShowGrid(false);
    disassemble();

    disassemblerTable->clearSelection();
    disassemblerTable->selectRow(MAXLINES / 2);

    // XXX: The next line does not work?? Why?
    disassemblerTable->setCurrentCell((MAXLINES / 2), 0);
    connect(disassemblerTable, SIGNAL(selectionChanged()), 
            this, SLOT(resetSelection()));
  }

  // QDisassemblerPane::~QDisassemblerPane()
  // Nothing to do here.
  QDisassemblerPane::~QDisassemblerPane()
  { }

  // void QDisassemblerPane::resetSelection()
  // No selection is allowed by the user. The class displaying the output
  // should be rewritten, the QTable is not the best solution for this
  // problem.
  void QDisassemblerPane::resetSelection()
  {
    if (disassemblerTable->isRowSelected((MAXLINES / 2)))
      return;
    
    disassemblerTable->clearSelection();
    disassemblerTable->selectRow(MAXLINES / 2);
  }

  // void QDisassemblerPane::addressChanged(unsigned int address)
  // This slot is used if the memory has changed.
  void QDisassemblerPane::addressChanged(unsigned int address)
  {
    int naddr = static_cast<int>(address);
    if (naddr > 32767)
      naddr -= 0x10000;
    int cpu_ip = static_cast<int>(cpu.getInstructionPointer());
    if (cpu_ip > 32767)
      cpu_ip -= 0x10000;

    if ((naddr < ((cpu_ip + MAXLINES))) &&
        (naddr > ((cpu_ip - MAXLINES))))
      disassemble();
  }

  // void QDisassemblerPane::disassemble()
  // Disassembles the memory around the current instruction pointer.
  void QDisassemblerPane::disassemble()
  {
    Interpreter::Decoder &decoder = cpu.getDecoder();
    std::string addr, words, opcode, arg;
    int ip = cpu.getInstructionPointer();
    int offset = 0, n;
    QStringList rowLabels;

    // first the memory after the instruction pointer
    for (n = (MAXLINES / 2); n < MAXLINES; n++) {
      // decode the memory
      int len = decoder.decode(((ip + offset)& 0xffff),
                               addr, words, opcode, arg);
      (disassemblerTable->item(n, 0))->setText(words.c_str());
      (disassemblerTable->item(n, 1))->setText(opcode.c_str());
      (disassemblerTable->item(n, 2))->setText(arg.c_str());
      // Add this address to the label list
      rowLabels.push_back(addr.c_str());
      offset += len;
    }

    offset = 0;
    for (n = (MAXLINES / 2) - 1; n >= 0; n--) {
      int len = decoder.decode(((ip - offset - 2) & 0xffff),
                               addr, words, opcode, arg);
      if (len == 1)
        len = decoder.decode(((ip - offset - 1) & 0xffff),
                             addr, words, opcode, arg);
      disassemblerTable->setText(n, 0, words.c_str());
      disassemblerTable->setText(n, 1, opcode.c_str());
      disassemblerTable->setText(n, 2, arg.c_str());
      // Add this address to the label list
      rowLabels.push_front(addr.c_str());

      offset += len;
    }
    
    // Now set the new labels
    disassemblerTable->setRowLabels(rowLabels);
  }
}
