/* $Id: QHexLineEdit.cpp,v 1.3 2004/11/21 20:23:20 markus Exp $ */

/*
 * Copyright (c) 2003 Markus Hennecke
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the Author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,

 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * The implementation of the HexLineEdit Widget. It has a Button that shows
 * if it is signed or unsigned before the LineEdit Widget and a Button
 * defineing the base after the LineEdit Widget. Base is either 16 (Hex)
 * or 10 (Dec). The input to the Widget is validated according to the two
 * additional buttons. The range must include zero because the input validation
 * is strict, no intermediate values may be accepted.
 */

// system includes:
#include <qpushbutton.h>
#include <qstring.h>
#include <qlayout.h>
#include <qlineedit.h>
#include <qtooltip.h>

// program includes:
#include "QHexLineEdit.hpp"
#include "QDecIntValidator.hpp"
#include "QHexIntValidator.hpp"

namespace GUI {

  //QHexLineEdit::QHexLineEdit(QWidget *parent, const char *name)
  //  : QWidget(parent, name), sign(false), base(16), number(0)
  // The constructor of the class. See the header file for an ASCII-art
  // image of how the widgets are arranged.
  QHexLineEdit::QHexLineEdit(QWidget *parent, const char *name)
    : QWidget(parent, name), sign(false), base(16), number(0)
  {
    boxLayout = new QHBoxLayout(this);
    signChoiceButton = new QPushButton(this);
    signChoiceButton->setText(tr("u"));
    signChoiceButton->setDisabled(true);
    QToolTip::add(signChoiceButton, tr("Toggle between signed and unsigned"));
    boxLayout->addWidget(signChoiceButton);

    numberHexValidator = new QHexIntValidator(this);
    numberHexValidator->setRange(0, 65535);
    numberDecValidator = new QDecIntValidator(this);
    numberDecValidator->setRange(0, 65535);

    numberLineEdit = new QLineEdit(this);
    numberLineEdit->setValidator(numberHexValidator);
    numberLineEdit->setAlignment(Qt::AlignRight);
    QToolTip::add(numberLineEdit, tr("The number to edit"));
    boxLayout->addWidget(numberLineEdit);

    baseChoiceButton = new QPushButton(this);
    baseChoiceButton->setText(tr("Hex"));
    QToolTip::add(baseChoiceButton, tr("Toggle between Hex- and Decimal"));
    boxLayout->addWidget(baseChoiceButton);

    // connect the signals to the slots
    connect(baseChoiceButton, SIGNAL(clicked()),
            this, SLOT(toggleBase()));
    connect(signChoiceButton, SIGNAL(clicked()),
            this, SLOT(toggleSignedUnsigned()));
    connect(numberLineEdit, SIGNAL(textChanged(const QString &)),
            this, SLOT(changed(const QString &)));
    connect(numberLineEdit, SIGNAL(returnPressed()),
            this, SLOT(retPressed()));
  }

  // QHexLineEdit::~QHexLineEdit()
  // The destructor. For now nothing to do.
  QHexLineEdit::~QHexLineEdit()
  { }

  // void QHexLineEdit::setNumber(int num)
  // Sets the number according to the limitations of a signed or unsigned
  // display and the base. The number is checked if it is in range and
  // adjusted if necessary.
  void QHexLineEdit::setNumber(int num)
  {
    // Check if the number can be displayed and adjust it if necessary
    if (base == 10) {
      if ((sign) && (num > 32767))
	num -= 65536;
      if ((! sign) && (num < 0))
	num += 65536;
    } else if (num < 0) {
      num += 65536;
    }

    // is it still in range??
    if ((num > 65535) || (num < -32768))
      return;

    // Print the number to the widget
    QString s;
    s.setNum(num, base);
    numberLineEdit->setText(s);
  }

  // void QHexLineEdit::toggleSignedUnsigned()
  // Toggles the signed / unsigned state of the widget. The number is 
  // adjusted to accommodate the new behaviour.
  // Note: Toggle of signed/unsigned display is only possible in decimal
  //       display.
  void QHexLineEdit::toggleSignedUnsigned()
  {
    if (sign) {
      sign = false;
      signChoiceButton->setText(tr("u"));
      numberDecValidator->setRange(0, 65535);
    } else {
      sign = true;
      signChoiceButton->setText(tr("s"));
      numberDecValidator->setRange(-32768, 32767);
    }

    // Modify a signed value if unsigned and vice versa
    if (sign) {
      if (number > 32767) {
        number -= 65536;
      }
    } else {
      if (number < 0) {
        number += 65536;
      }
    }

    // Set the text on the widget so the adjusted behaviour is displayed
    QString s;

    s.setNum(number, base);
    numberLineEdit->setText(s);
  }

  // void QHexLineEdit::toggleBase()
  // Changes the base of the number displayed from Hex to Dec and vice versa.
  // TODO: This could be extended to support Dual numbers as well. Note that
  // the maximum input length of the LineEdit Widget must be adjusted too to
  // enable full 16bit input.
  void QHexLineEdit::toggleBase()
  {
    // hm, this can be left out since textChanged takes care of it??
    //bool ok;
    // get the number from the widget...
    //number = numberLineEdit->text().toInt(&ok, base);

    if (base != 10) {
      base = 10;
      baseChoiceButton->setText(tr("Dec"));
      if (sign)
        numberDecValidator->setRange(-32768, 32767);
      numberLineEdit->setValidator(numberDecValidator);
      signChoiceButton->setDisabled(false);
      if (sign)
	signChoiceButton->setText(tr("s"));
      else
	signChoiceButton->setText(tr("u"));
      setNumber(number);
    } else {
      base = 16;
      baseChoiceButton->setText(tr("Hex"));
      numberHexValidator->setRange(0, 65535);
      numberLineEdit->setValidator(numberHexValidator);
      signChoiceButton->setDisabled(true);
      signChoiceButton->setText(tr("u"));
      if (number < 0)
	number += 65536;
    }

    // Set the new base to the widgets display
    QString s;
    s.setNum(number, base);
    numberLineEdit->setText(s);
  }

  // void QHexLineEdit::changed(const QString &)
  // If the text changes the new number must be read from the LineEdit Widget.
  void QHexLineEdit::changed(const QString &s)
  {
    // an empty string defaults to 0
    if (s.isEmpty()) {
      number = 0;
    } else {
      bool ok;
      number = s.toInt(&ok, base);
    }
  } 

  // void QHexLineEdit::retPressed()
  // Dummy slot to simulate the behaviour of a QLineEdit Widget.
  void QHexLineEdit::retPressed()
  {
    emit(returnPressed());
  }

}
