/* $Id: QMemoryPane.cpp,v 1.4 2003/12/12 09:47:27 markus Exp $ */

/*
 * Copyright (c) 2003 Markus Hennecke
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the Author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,

 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * The implementation of a Widget that contains a MemoryTable, a chooser to set
 * the base of the MemoryTable and a HexLineEdit Widget to edit a selection.
 */

// system includes:
#include <qpushbutton.h>
#include <qlayout.h>
#include <qcombobox.h>
#include <qlineedit.h>
#include <qtooltip.h>

// program includes:
#include "QMemoryPane.hpp"

namespace GUI {

  // QMemoryPane::QMemoryPane(QWidget *parent, const char *name)
  //   : QWidget(parent, name)
  // Constructor that constructs the widgets child widgets and sets everything
  // up.
  QMemoryPane::QMemoryPane(QWidget *parent, const char *name)
    : QWidget(parent, name)
  {
    // We have a vertical main Layout.
    QVBoxLayout *mainLayout = new QVBoxLayout(this);
    // Don't give a MemoryChain to the memTable yet
    memTable = new QMemoryTable(NULL, this);
    mainLayout->addWidget(memTable);

    // The next Layout holds all the buttons and edit fields
    QHBoxLayout *buttonLayout = new QHBoxLayout(mainLayout);

    memoryRangeComboBox = new QComboBox(this);
    memoryRangeComboBox->insertItem(tr("0000 - 3FFF"));
    memoryRangeComboBox->insertItem(tr("4000 - 7FFF"));
    memoryRangeComboBox->insertItem(tr("8000 - BFFF"));
    memoryRangeComboBox->insertItem(tr("C000 - FFFF"));
    QToolTip::add(memoryRangeComboBox,
                  tr("Choose a range of memory to display"));
    buttonLayout->addWidget(memoryRangeComboBox);

    // Setup the Edit field
    memoryLineEdit = new QHexLineEdit(this);
    memoryLineEdit->setEnabled(false);
    buttonLayout->addWidget(memoryLineEdit);

    changeValueButton = new QPushButton(tr("Accept value"), this);
    changeValueButton->setEnabled(false);
    QToolTip::add(changeValueButton,
                  tr("Commit the changes"));
    buttonLayout->addWidget(changeValueButton);

    // notify us if the user did something
    connect(memoryRangeComboBox, SIGNAL(highlighted(int)), 
            this, SLOT(comboChanged(int)));
    connect(memTable, SIGNAL(select(int, int)), 
            this, SLOT(memorySelected(int, int)));
    connect(changeValueButton, SIGNAL(clicked()),
            this, SLOT(memoryEdited()));
    connect(memoryLineEdit, SIGNAL(returnPressed()),
            this, SLOT(memoryEdited()));
  }

  // QMemoryPane::~QMemoryPane()
  // Nothing to do here.
  QMemoryPane::~QMemoryPane()
  { }

  // void QMemoryPane::comboChanged(int)
  // Sets a new Base to the QMemoryTable Widget if the user has changed
  // the selection.
  void QMemoryPane::comboChanged(int index)
  {
    // calculate the new base
    unsigned int newBase = index * 0x4000;

    // check if the new base is really new
    if (memTable->getBase() == newBase)
      return;

    // if it is new set the new base
    memTable->setBase(index * 0x4000);

    // the QMemoryTable resets all selections so the widgets must be disabled
    changeValueButton->setEnabled(false);
    memoryLineEdit->setEnabled(false);
  }
  
  // void QMemoryPane::setMemory(Interpreter::MemoryChain *)#
  // Sets a new MemoryChain to be displayed.
  void QMemoryPane::setMemory(Interpreter::MemoryChain *mem)
  {
    memTable->setMem(mem);
  }

  // void QMemoryPane::memorySelected(int row, int col)
  // This slot is used if a Selection occured in the QMemoryTable Widget.
  // The widgets to edit a selected memory cell are (de)-activated accordingly
  // to the state of the selection.
  void QMemoryPane::memorySelected(int row, int col)
  {
    if (! ((row == -1) && (col == -1))) {
      // Do we have mem?
      Interpreter::MemoryChain *mem = memTable->getMem();
      if (mem) {
        // yes, then enable edits.
        short number;
        unsigned int address = memTable->getBase() + (row * 8) + col - 1;
        if (mem->read(address, number)) {
          changeValueButton->setEnabled(true);
          memoryLineEdit->setEnabled(true);
          memoryLineEdit->setNumber(number);
          return;
        }
      }
    }
    // nothing selected, or an invalid field. disable any edit facililties
    changeValueButton->setEnabled(false);
    memoryLineEdit->setEnabled(false);    
  }

  // void QMemoryPane::memoryEdited() slot
  // Either return is pressed in the HexLineEdit Widget or the Button 
  // Commit change was clicked. The value has to be written to memory and
  // the changed value has to be displayed.
  void QMemoryPane::memoryEdited()
  {
    int newval = memoryLineEdit->getNumber();

    // adjust it so it will fit in a short
    if (newval > 32767)
      newval -= 65536;

    memTable->changeSelection(static_cast<short>(newval));
    unsigned int address = memTable->getBase() + 
      (memTable->getSelectedRow() * 8) + memTable->getSelectedCol() - 1;
    emit(memoryChanged(address));
  }

  // void QMemoryPane::memoryWritten(unsigned int) slot
  // Notify the display that there was a write operation to memory that
  // has to be (possibly) displayed.
  void QMemoryPane::memoryWritten(unsigned int address)
  {
    memTable->changedMem(address);
  }

}
