/* $Id: QMemoryTable.cpp,v 1.4 2003/12/12 09:47:27 markus Exp $ */

/*
 * Copyright (c) 2003 Markus Hennecke
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the Author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,

 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * The implementation of a fast and simple memory Tableview. This one displays
 * 16k words of memory in rows of 8 words. At the beginning of each row
 * the current address is printed, with a grey background. Every cell can
 * be selected, except the address cells. There is no possibility of selecting
 * more than one cell.
 */

// system includes:
#include <qtooltip.h>
#include <qpainter.h>

// program includes:
#include "QMemoryTable.hpp"

namespace GUI {

  // QMemoryTable::QMemoryTable(Interpreter::MemoryChain *mem, QWidget *parent,
  //                            const char *name, WFlags f)
  //   : QGridView(parent, name, f), mem(mem), base(0), selectedRow(-1), 
  //     selectedCol(-1)
  // The Constructor initializes the properties and tries to guess
  // a proper cell Geometry.
  QMemoryTable::QMemoryTable(Interpreter::MemoryChain *mem, QWidget *parent,
                             const char *name, WFlags f)
    : QGridView(parent, name, f), mem(mem), base(0), selectedRow(-1), 
      selectedCol(-1)
  {
    setFont(QFont("fixed"));
    QFontMetrics fm(font());

    // this is a hack to get a somewhat usefull cellwidth
    pixelWidth = fm.width("CCCC:");
    setCellWidth(pixelWidth + 6);
    pixelHeight = fm.height();
    setCellHeight(pixelHeight + 4);

    setNumCols(9);    // we have the address and 8 words after the address
    setNumRows(2048); // 8192 rows in 4 parts
    
    QToolTip::add(this, 
           tr("The content of the memory.\nClick to select a specific cell"));
  }

  // QMemoryTable::~QMemoryTable()
  // The destructor. Nothing to be done here.
  QMemoryTable::~QMemoryTable()
  { }

  // void QMemoryTable::paintCell(QPainter *, int, int)
  // Paints the cell given by row and col at the region given by
  // the QPainter object.
  void QMemoryTable::paintCell(QPainter *p, int row, int col)
  {
    QString s;
    short val;
    
    // get the cell Geometry
    QRect rect(cellRect());

    // Save the brush, backgroundColor and Pen
    const QColor bgcolor(p->backgroundColor());
    const QBrush bgbrush(p->brush());
    const QPen fgpen(p->pen());
    BGMode bgmode = p->backgroundMode();

    // If there is no memory at the given address -1 (0xffff) in light grey
    // color will be painted.
    bool is_mem = false;

    if (col) {

      // Paint a memory cell
      if (mem) {
        // check if there is some memory at the address
        is_mem = mem->read(base + (8 * row) + col - 1, val);
      } else {
        val = -1;
      }

      // print a value from the given address
      s.sprintf("%04X", val & 0xffff);
      QFontMetrics fm(font());
      int width = fm.width(s);

      if ((col == selectedCol) && (row == selectedRow)) {
        // is it selected??
        p->setBackgroundMode(Qt::OpaqueMode);
        p->setPen(Qt::blue);
        p->setBrush(Qt::blue);
        p->setBackgroundColor(Qt::blue);
        p->drawRect(rect);
        p->setPen(Qt::white);
        p->drawText(((rect.width() - width) / 2), pixelHeight, s, -1);
      } else {
        // check if there was memory, if not paint it light grey
        if (! is_mem) {
          p->setPen(QColor(200, 200, 200));
        }
        p->drawText(((rect.width() - width) / 2), pixelHeight, s, -1);
      }

    } else {

      // print the address with a grey background and a black line at the
      // right of the cell
      val = base + (8 * row);
      s.sprintf("%04X:", val & 0xffff);
      QColor grey(190, 190, 190);
      p->setBrush(grey);
      p->setBackgroundColor(grey);
      p->setPen(grey);
      p->drawRect(rect);
      p->setPen(fgpen);
      p->drawLine(rect.right(), rect.bottom(), rect.right(), rect.top());
      p->drawText(2, pixelHeight + 2, s, -1);
    }

    // set everything back to the saved values
    p->setPen(fgpen);
    p->setBackgroundColor(bgcolor);
    p->setBackgroundMode(bgmode);
  }

  // void QMemoryTable::contentsMousePressEvent(QMouseEvent *)
  // Handles a mouse press event. Normaly this is a selection or a
  // previously selected cell is deselected.
  void QMemoryTable::contentsMousePressEvent(QMouseEvent *e)
  {
    // save the old selection
    int oldRow = selectedRow;
    int oldCol = selectedCol;

    // get the new selection
    selectedRow = rowAt(e->y());
    selectedCol = columnAt(e->x());

    if (selectedCol == 0) {
      // clicked on the address?? the return imm.
      selectedRow = oldRow;
      selectedCol = oldCol;
    } else if ((oldRow == selectedRow) && (oldCol == selectedCol)) {
      // look if it was the same?
      selectedRow = -1;
      selectedCol = -1;
      updateCell(oldRow, oldCol);
    } else {
      updateCell(selectedRow, selectedCol);
      updateCell(oldRow, oldCol);
    }

    // emit the signal
    emit(select(selectedRow, selectedCol));
  }

  // void QMemoryTable::changeSelection(const short newval)
  // called if the memory was changed and a repaint has to be done.
  void QMemoryTable::changeSelection(const short newval)
  {
    // Only do something if a selection is active
    if ((selectedRow != -1) && (selectedCol != -1)) {
      unsigned int address = base + (selectedRow * 8) + selectedCol - 1;
      mem->write(address, newval);
      updateCell(selectedRow, selectedCol);
    }
  }

  // void QMemoryTable::changedMem(unsigned int address)
  // Paints the selected address if it is visible.
  void QMemoryTable::changedMem(unsigned int address)
  {
    // may the address be visible?
    if (address >= base + 0x4000)
      return;

    int row = (address - base) / 8;
    int col = ((address - base) % 8) + 1;
    updateCell(row, col);
  }
}
