/* $Id: CPU.cpp,v 1.9 2004/11/21 20:52:49 markus Exp $ */

/*
 * Copyright (c) 2003 Markus Hennecke
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the Author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Implementation of the virtual cpu.
 */

// system includes:
#ifndef GUI_ONLY
#include <iomanip>
#endif

// program includes:
#include "CPU.hpp"
#include "Error.hpp"

namespace Interpreter {

  // void CPU::hlt(const Decoder::AddressMode mode, const int arg) 
  //   throw (Error::illegal_instruction)
  void CPU::hlt(const Decoder::AddressMode mode) 
    throw (Error::illegal_instruction)
  {
    if (mode != Decoder::NONE)
      throw Error::illegal_instruction(InstructionPointer);
    // don't step any further, but halt the cpu
    jump = true;
    halt(true);
  }

  // void CPU::lad(const Decoder::AddressMode mode, const int arg)
  //   throw (Error::illegal_instruction)
  void CPU::lad(const Decoder::AddressMode mode, const int arg)
    throw (Error::illegal_instruction)
  {
    // some helper variables for memory arithmetics...
    short read_val;
    unsigned int ind_addr;

    switch (mode) {
    case Decoder::DIRECT:
      Akku = static_cast<short>(arg);
      break;
    case Decoder::ADDRESS:
      memChain.read(static_cast<unsigned int>(arg), Akku);
      break;
    case Decoder::RELATIVE:
      memChain.read(static_cast<unsigned int>(arg + InstructionPointer), Akku);
      break;
    case Decoder::INDIRECT:
      memChain.read(static_cast<unsigned int>(arg), read_val);
      ind_addr = (read_val < 0) ? read_val+65536 : read_val;
      memChain.read(ind_addr, Akku);
      break;
    case Decoder::INDEX:
      memChain.read(static_cast<unsigned int>(arg + Index), Akku);
      break;
    default:
      throw Error::illegal_instruction(InstructionPointer);
    }
  }

  // void CPU::spi(const Decoder::AddressMode mode, const int arg)
  //   throw (Error::illegal_instruction)
  void CPU::spi(const Decoder::AddressMode mode, const int arg)
    throw (Error::illegal_instruction)
  {
    // some helper variables for memory arithmetics...
    short read_val;

    switch (mode) {
    case Decoder::ADDRESS:
      lastWriteAddress = static_cast<unsigned int>(arg);
      memChain.write(lastWriteAddress, Akku);
      break;
    case Decoder::RELATIVE:
      lastWriteAddress = static_cast<unsigned int>(arg + InstructionPointer);
      memChain.write(lastWriteAddress, Akku);
      break;
    case Decoder::INDIRECT:
      memChain.read(static_cast<unsigned int>(arg), read_val);
      lastWriteAddress = (read_val < 0) ? read_val+65536 : read_val;
      memChain.write(lastWriteAddress, Akku);
      break;
    case Decoder::INDEX:
      lastWriteAddress = static_cast<unsigned int>(arg + Index);
      memChain.write(lastWriteAddress, Akku);
      break;
    default:
      throw Error::illegal_instruction(InstructionPointer);
    }
  }

  // void CPU::add(const Decoder::AddressMode mode, const int arg)
  //   throw (Error::illegal_instruction)
  void CPU::add(const Decoder::AddressMode mode, const int arg)
    throw (Error::illegal_instruction)
  {
    // some helper variables for memory arithmetics...
    short read_val;
    unsigned int ind_addr;

    switch (mode) {
    case Decoder::DIRECT:
      read_val = static_cast<short>(arg);
      break;
    case Decoder::ADDRESS:
      memChain.read(static_cast<unsigned int>(arg), read_val);
      break;
    case Decoder::RELATIVE:
      memChain.read(static_cast<unsigned int>(arg + InstructionPointer), 
                    read_val);
      break;
    case Decoder::INDIRECT:
      memChain.read(static_cast<unsigned int>(arg), read_val);
      ind_addr = (read_val < 0) ? read_val+65536 : read_val;
      memChain.read(ind_addr, read_val);
      break;
    case Decoder::INDEX:
      memChain.read(static_cast<unsigned int>(arg + Index), read_val);
      break;
    default:
      throw Error::illegal_instruction(InstructionPointer);
    }
    Akku += read_val;
  }

  // void CPU::sub(const Decoder::AddressMode mode, const int arg)
  //   throw (Error::illegal_instruction)
  void CPU::sub(const Decoder::AddressMode mode, const int arg)
    throw (Error::illegal_instruction)
  {
    // some helper variables for memory arithmetics...
    short read_val;
    unsigned int ind_addr;

    switch (mode) {
    case Decoder::DIRECT:
      read_val = static_cast<short>(arg);
      break;
    case Decoder::ADDRESS:
      memChain.read(static_cast<unsigned int>(arg), read_val);
      break;
    case Decoder::RELATIVE:
      memChain.read(static_cast<unsigned int>(arg + InstructionPointer), 
                    read_val);
      break;
    case Decoder::INDIRECT:
      memChain.read(static_cast<unsigned int>(arg), read_val);
      ind_addr = (read_val < 0) ? read_val+65536 : read_val;
      memChain.read(ind_addr, read_val);
      break;
    case Decoder::INDEX:
      memChain.read(static_cast<unsigned int>(arg + Index), read_val);
      break;
    default:
      throw Error::illegal_instruction(InstructionPointer);
    }
    Akku -= read_val;
  }

  // void CPU::mul(const Decoder::AddressMode mode, const int arg)
  //   throw (Error::illegal_instruction)
  void CPU::mul(const Decoder::AddressMode mode, const int arg)
    throw (Error::illegal_instruction)
  {
    // some helper variables for memory arithmetics...
    short read_val;
    unsigned int ind_addr;

    switch (mode) {
    case Decoder::DIRECT:
      read_val = static_cast<short>(arg);
      break;
    case Decoder::ADDRESS:
      memChain.read(static_cast<unsigned int>(arg), read_val);
      break;
    case Decoder::RELATIVE:
      memChain.read(static_cast<unsigned int>(arg + InstructionPointer), 
                    read_val);
      break;
    case Decoder::INDIRECT:
      memChain.read(static_cast<unsigned int>(arg), read_val);
      ind_addr = (read_val < 0) ? read_val+65536 : read_val;
      memChain.read(ind_addr, read_val);
      break;
    case Decoder::INDEX:
      memChain.read(static_cast<unsigned int>(arg + Index), read_val);
      break;
    default:
      throw Error::illegal_instruction(InstructionPointer);
    }
    // TODO: what happens if an overflow occurs??
    Akku *= read_val;
  }

  // void CPU::div(const Decoder::AddressMode mode, const int arg)
  //   throw (Error::illegal_instruction)
  void CPU::div(const Decoder::AddressMode mode, const int arg)
    throw (Error::illegal_instruction, Error::division_by_zero)
  {
    // some helper variables for memory arithmetics...
    short read_val;
    unsigned int ind_addr;

    switch (mode) {
    case Decoder::DIRECT:
      read_val = static_cast<short>(arg);
      break;
    case Decoder::ADDRESS:
      memChain.read(static_cast<unsigned int>(arg), read_val);
      break;
    case Decoder::RELATIVE:
      memChain.read(static_cast<unsigned int>(arg + InstructionPointer), 
                    read_val);
      break;
    case Decoder::INDIRECT:
      memChain.read(static_cast<unsigned int>(arg), read_val);
      ind_addr = (read_val < 0) ? read_val+65536 : read_val;
      memChain.read(ind_addr, read_val);
      break;
    case Decoder::INDEX:
      memChain.read(static_cast<unsigned int>(arg + Index), read_val);
      break;
    default:
      throw Error::illegal_instruction(InstructionPointer);
    }
    // TODO: what happens if an overflow occurs??
    
    if (! read_val) {
      halt(true);
      throw Error::division_by_zero(InstructionPointer);
    }

    Help = Akku % read_val;
    Akku /= read_val;
  }

  // void CPU::jmp(const Decoder::AddressMode mode, const int arg)
  //   throw (Error::illegal_instruction)
  void CPU::jmp(const Decoder::AddressMode mode, const int arg)
    throw (Error::illegal_instruction)
  {
    // some helper variables for memory arithmetics...
    short read_val;
    unsigned int ind_addr;

    switch (mode) {
    case Decoder::ADDRESS:
      ind_addr = static_cast<unsigned int>(arg);
      break;
    case Decoder::RELATIVE:
      ind_addr = static_cast<unsigned int>(arg + InstructionPointer);
      break;
    case Decoder::INDIRECT:
      memChain.read(static_cast<unsigned int>(arg), read_val);
      ind_addr = (read_val < 0) ? read_val+65536 : read_val;
      break;
    case Decoder::INDEX:
      ind_addr = static_cast<unsigned int>(arg + Index);
      break;
    default:
      throw Error::illegal_instruction(InstructionPointer);
    }
    InstructionPointer = static_cast<unsigned int>(ind_addr);
    jump = true;
  }

  // void CPU::jez(const Decoder::AddressMode mode, const int arg)
  //   throw (Error::illegal_instruction)
  void CPU::jez(const Decoder::AddressMode mode, const int arg)
    throw (Error::illegal_instruction)
  {
    // some helper variables for memory arithmetics...
    short read_val;
    unsigned int ind_addr;

    switch (mode) {
    case Decoder::ADDRESS:
      ind_addr = static_cast<unsigned int>(arg);
      break;
    case Decoder::RELATIVE:
      ind_addr = static_cast<unsigned int>(arg + InstructionPointer);
      break;
    case Decoder::INDIRECT:
      memChain.read(static_cast<unsigned int>(arg), read_val);
      ind_addr = (read_val < 0) ? read_val+65536 : read_val;
      break;
    case Decoder::INDEX:
      ind_addr = static_cast<unsigned int>(arg + Index);
      break;
    default:
      throw Error::illegal_instruction(InstructionPointer);
    }
    if (Akku == 0) {
      InstructionPointer = static_cast<unsigned int>(ind_addr);
      jump = true;
    }
  }

  // void CPU::jgz(const Decoder::AddressMode mode, const int arg)
  //   throw (Error::illegal_instruction)
  void CPU::jgz(const Decoder::AddressMode mode, const int arg)
    throw (Error::illegal_instruction)
  {
    // some helper variables for memory arithmetics...
    short read_val;
    unsigned int ind_addr;

    switch (mode) {
    case Decoder::ADDRESS:
      ind_addr = static_cast<unsigned int>(arg);
      break;
    case Decoder::RELATIVE:
      ind_addr = static_cast<unsigned int>(arg + InstructionPointer);
      break;
    case Decoder::INDIRECT:
      memChain.read(static_cast<unsigned int>(arg), read_val);
      ind_addr = (read_val < 0) ? read_val+65536 : read_val;
      break;
    case Decoder::INDEX:
      ind_addr = static_cast<unsigned int>(arg + Index);
      break;
    default:
      throw Error::illegal_instruction(InstructionPointer);
    }
    if (Akku >= 0) {
      InstructionPointer = static_cast<unsigned int>(ind_addr);
      jump = true;
    }
  }

  // void CPU::jlz(const Decoder::AddressMode mode, const int arg)
  //   throw (Error::illegal_instruction)
  void CPU::jlz(const Decoder::AddressMode mode, const int arg)
    throw (Error::illegal_instruction)
  {
    // some helper variables for memory arithmetics...
    short read_val;
    unsigned int ind_addr;

    switch (mode) {
    case Decoder::ADDRESS:
      ind_addr = static_cast<unsigned int>(arg);
      break;
    case Decoder::RELATIVE:
      ind_addr = static_cast<unsigned int>(arg + InstructionPointer);
      break;
    case Decoder::INDIRECT:
      memChain.read(static_cast<unsigned int>(arg), read_val);
      ind_addr = (read_val < 0) ? read_val+65536 : read_val;
      break;
    case Decoder::INDEX:
      ind_addr = static_cast<unsigned int>(arg + Index);
      break;
    default:
      throw Error::illegal_instruction(InstructionPointer);
    }
    if (Akku < 0) {
      InstructionPointer = static_cast<unsigned int>(ind_addr);
      jump = true;
    }
  }

  // void CPU::swap(const Decoder::AddressMode mode)
  //   throw (Error::illegal_instruction)
  void CPU::swap(const Decoder::AddressMode mode)
    throw (Error::illegal_instruction)
  {
    if (mode != Decoder::NONE)
      throw Error::illegal_instruction(InstructionPointer);
    short tmp = Akku;
    Akku = Help;
    Help = tmp;
  }

  // void CPU::jsr(const Decoder::AddressMode mode, const int arg)
  //   throw (Error::illegal_instruction)
  void CPU::jsr(const Decoder::AddressMode mode, const int arg)
    throw (Error::illegal_instruction)
  {
    // some helper variables for memory arithmetics...
    short read_val;
    unsigned int ind_addr;

    switch (mode) {
    case Decoder::ADDRESS:
      ind_addr = static_cast<unsigned int>(arg);
      break;
    case Decoder::RELATIVE:
      ind_addr = static_cast<unsigned int>(arg + InstructionPointer);
      break;
    case Decoder::INDIRECT:
      memChain.read(static_cast<unsigned int>(arg), read_val);
      ind_addr = (read_val < 0) ? read_val+65536 : read_val;
      break;
    case Decoder::INDEX:
      ind_addr = static_cast<unsigned int>(arg + Index);
      break;
    default:
      throw Error::illegal_instruction(InstructionPointer);
    }
    push(Decoder::DIRECT, InstructionPointer + len);
    InstructionPointer = static_cast<unsigned int>(ind_addr);
    jump = true;
  }

  // void CPU::ret(const Decoder::AddressMode mode, const int arg)
  //   throw (Error::illegal_instruction)
  void CPU::ret(const Decoder::AddressMode mode)
    throw (Error::illegal_instruction)
  {
    if (mode != Decoder::NONE)
      throw Error::illegal_instruction(InstructionPointer);

    short read_val;
    ++StackPointer;
    StackPointer &= 0xffff;

    memChain.read(StackPointer, read_val);

    InstructionPointer = static_cast<unsigned int>(
               (read_val < 0) ? read_val+65536 : read_val);
    jump = true;
  }

  // void CPU::lsa(const Decoder::AddressMode mode)
  //   throw (Error::illegal_instruction)
  void CPU::lsa(const Decoder::AddressMode mode)
    throw (Error::illegal_instruction)
  {
    if (mode != Decoder::NONE)
      throw Error::illegal_instruction(InstructionPointer);

    StackPointer = static_cast<unsigned int>((Akku < 0) ? Akku+65536 : Akku);
  }

  // void CPU::lds(const Decoder::AddressMode mode)
  //   throw (Error::illegal_instruction)
  void CPU::lds(const Decoder::AddressMode mode)
    throw (Error::illegal_instruction)
  {
    if (mode != Decoder::NONE)
      throw Error::illegal_instruction(InstructionPointer);

    Akku = static_cast<short>((StackPointer > 32767) ? StackPointer-65535 : StackPointer);

  }

  // void CPU::cli(const Decoder::AddressMode mode)
  //   throw (Error::illegal_instruction)
  void CPU::cli(const Decoder::AddressMode mode)
    throw (Error::illegal_instruction)
  {
    if (mode != Decoder::NONE)
      throw Error::illegal_instruction(InstructionPointer);

    Index = 0;
  }

  // void CPU::ldi(const Decoder::AddressMode mode)
  //   throw (Error::illegal_instruction)
  void CPU::ldi(const Decoder::AddressMode mode)
    throw (Error::illegal_instruction)
  {
    if (mode != Decoder::NONE)
      throw Error::illegal_instruction(InstructionPointer);

    Index = Akku;
  }

  // void CPU::lia(const Decoder::AddressMode mode, const int arg)
  //   throw (Error::illegal_instruction)
  void CPU::lia(const Decoder::AddressMode mode)
    throw (Error::illegal_instruction)
  {
    if (mode != Decoder::NONE)
      throw Error::illegal_instruction(InstructionPointer);

    Akku = Index;
  }

  // void CPU::ink(const Decoder::AddressMode mode, const int arg)
  //   throw (Error::illegal_instruction)
  void CPU::ink(const Decoder::AddressMode mode)
    throw (Error::illegal_instruction)
  {
    if (mode != Decoder::NONE)
      throw Error::illegal_instruction(InstructionPointer);

    Index++;
    Index &= 0xffff;
  }

  // void CPU::dek(const Decoder::AddressMode mode, const int arg)
  //   throw (Error::illegal_instruction)
  void CPU::dek(const Decoder::AddressMode mode)
    throw (Error::illegal_instruction)
  {
    if (mode != Decoder::NONE)
      throw Error::illegal_instruction(InstructionPointer);

    Index--;
    Index &= 0xffff;
  }

  // void CPU::ldx(const Decoder::AddressMode mode, const int arg)
  //   throw (Error::illegal_instruction)
  void CPU::ldx(const Decoder::AddressMode mode, const int arg)
    throw (Error::illegal_instruction)
  {
    if (mode != Decoder::ADDRESS)
      throw Error::illegal_instruction(InstructionPointer);

    short newIndex;
    memChain.read(static_cast<unsigned int>(arg), newIndex);
    Index = static_cast<unsigned int>((newIndex < 0) ? newIndex+65536 : newIndex);
  }

  // void CPU::pushac(const Decoder::AddressMode mode)
  //   throw (Error::illegal_instruction)
  void CPU::pushac(const Decoder::AddressMode mode)
    throw (Error::illegal_instruction)
  {
    if (mode != Decoder::NONE)
      throw Error::illegal_instruction(InstructionPointer);

    lastWriteAddress = StackPointer;
    memChain.write(StackPointer--, Akku);
    StackPointer &= 0xffff;
  }

  // void CPU::popac(const Decoder::AddressMode mode, const int arg)
  //   throw (Error::illegal_instruction)
  void CPU::popac(const Decoder::AddressMode mode)
    throw (Error::illegal_instruction)
  {
    if (mode != Decoder::NONE)
      throw Error::illegal_instruction(InstructionPointer);

    ++StackPointer;
    // prevent overflow
    StackPointer &= 0xffff;

    memChain.read(StackPointer, Akku);
  }

  // void CPU::addst(const Decoder::AddressMode mode)
  //   throw (Error::illegal_instruction)
  void CPU::addst(const Decoder::AddressMode mode)
    throw (Error::illegal_instruction)
  {
    if (mode != Decoder::NONE)
      throw Error::illegal_instruction(InstructionPointer);

    short op1, op2;
    ++StackPointer;
    // prevent overflow
    StackPointer &= 0xffff;

    memChain.read(StackPointer, op1);

    ++StackPointer;
    // prevent overflow
    StackPointer &= 0xffff;

    memChain.read(StackPointer, op2);
    lastWriteAddress = StackPointer;
    memChain.write(StackPointer--, (op1 + op2));
    StackPointer &= 0xffff;
  }

  // void CPU::subst(const Decoder::AddressMode mode)
  //   throw (Error::illegal_instruction)
  void CPU::subst(const Decoder::AddressMode mode)
    throw (Error::illegal_instruction)
  {
    if (mode != Decoder::NONE)
      throw Error::illegal_instruction(InstructionPointer);

    short op1, op2;
    ++StackPointer;
    // prevent overflow
    StackPointer &= 0xffff;

    memChain.read(StackPointer, op1);

    ++StackPointer;
    // prevent overflow
    StackPointer &= 0xffff;

    memChain.read(StackPointer, op2);
    lastWriteAddress = StackPointer;
    memChain.write(StackPointer--, (op1 - op2));
    StackPointer &= 0xffff;
  }

  // void CPU::mulst(const Decoder::AddressMode mode)
  //   throw (Error::illegal_instruction)
  void CPU::mulst(const Decoder::AddressMode mode)
    throw (Error::illegal_instruction)
  {
    if (mode != Decoder::NONE)
      throw Error::illegal_instruction(InstructionPointer);

    short op1, op2;
    ++StackPointer;
    // prevent overflow
    StackPointer &= 0xffff;

    memChain.read(StackPointer, op1);
    
    ++StackPointer;
    // prevent overflow
    StackPointer &= 0xffff;

    memChain.read(StackPointer, op2);
    lastWriteAddress = StackPointer;
    memChain.write(StackPointer--, (op1 * op2));
    StackPointer &= 0xffff;
  }

  // void CPU::divst(const Decoder::AddressMode mode)
  //   throw (Error::illegal_instruction)
  void CPU::divst(const Decoder::AddressMode mode)
    throw (Error::illegal_instruction, Error::division_by_zero)
  {
    if (mode != Decoder::NONE)
      throw Error::illegal_instruction(InstructionPointer);

    short op1, op2;
    ++StackPointer;
    // prevent overflow
    StackPointer &= 0xffff;

    memChain.read(StackPointer, op1);
    
    ++StackPointer;
    // prevent overflow
    StackPointer &= 0xffff;

    memChain.read(StackPointer, op2);
    lastWriteAddress = StackPointer;
    if (op2) {
      memChain.write(StackPointer--, (op1 / op2));
    } else {
      halt(true);
      throw Error::division_by_zero(InstructionPointer);
    }
    StackPointer &= 0xffff;
  }

  // void CPU::push(const Decoder::AddressMode mode, const int arg)
  //   throw (Error::illegal_instruction)
  void CPU::push(const Decoder::AddressMode mode, const int arg)
    throw (Error::illegal_instruction)
  {
    if (mode != Decoder::DIRECT)
      throw Error::illegal_instruction(InstructionPointer);

    lastWriteAddress = StackPointer;
    memChain.write(StackPointer--, static_cast<short>(arg));
    StackPointer &= 0xffff;
  }

  // void CPU::pop(const Decoder::AddressMode mode, const int arg)
  //   throw (Error::illegal_instruction)
  void CPU::pop(const Decoder::AddressMode mode, const int arg)
    throw (Error::illegal_instruction)
  {
    if (mode != Decoder::NONE)
      throw Error::illegal_instruction(InstructionPointer);

    short val;

    ++StackPointer;
    // prevent overflow
    StackPointer &= 0xffff;

    memChain.read(StackPointer, val);
    StackPointer &= 0xffff;

    lastWriteAddress = static_cast<unsigned int>(arg);
    memChain.write(lastWriteAddress, val);
  }


  // bool CPU::step() 
  //   throw (Error::illegal_instruction)
  // step decodes the operation stored at InstructionPointer with the help
  // of the decoder and schedules the appropriate action. If there was an
  // illegal bit combination at the Address InstructionPointer points to 
  // a Error::illegal_instruction exception will be thrown.
  bool CPU::step() 
    throw (Error::illegal_instruction, Error::division_by_zero)
  {
    Decoder::AddressMode mode;
    int arg, code;

    len = decoder.decode(InstructionPointer, mode, code, arg);
    switch (code) {

    case 0x00: // HLT
      hlt(mode);
      break;

    case 0x01: // LAD
      lad(mode, arg);
      break;

    case 0x02: // SPI
      spi(mode, arg);
      break;

    case 0x03: // ADD
      add(mode, arg);
      break;

    case 0x04: // SUB
      sub(mode, arg);
      break;

    case 0x05: // MUL
      mul(mode, arg);
      break;

    case 0x06: // DIV
      div(mode, arg);
      break;

    case 0x07: // JMP
      jmp(mode, arg);
      break;

    case 0x08: // JEZ
      jez(mode, arg);
      break;

    case 0x09: // JGZ
      jgz(mode, arg);
      break;

    case 0x0a: // JLZ
      jlz(mode, arg);
      break;

    case 0x0b: // SWAP
      swap(mode);
      break;

    case 0xc1: // PUSHAC
      pushac(mode);
      break;

    case 0xc2: // POPAC
      popac(mode);
      break;

    case 0xc3: // ADDST
      addst(mode);
      break;

    case 0xc4: // SUBST
      subst(mode);
      break;

    case 0xc5: // MULST
      mulst(mode);
      break;
      
    case 0xc6: // DIVST
      divst(mode);
      break;

    case 0xc7: // LSA
      lsa(mode);
      break;
   
    case 0xc8: // LDS
      lds(mode);
      break;

      
    case 0xcb: // PUSH
      // there are some illegal opcodes here that could be used for
      // some extended addressing mode for push. not implemented, so.
      push(mode, arg);
      break;

    case 0xcc: // POP
      // see the comments on PUSH
      pop(mode, arg);
      break;

    case 0xd: // JSR
      jsr(mode, arg);
      break;
 
    case 0xe: // RET
      ret(mode);
      break;

    case 0xf0: // CLI
      cli(mode);
      break;

    case 0xf1: // LDI
      ldi(mode);
      break;

    case 0xf2: // LIA
      lia(mode);
      break;

    case 0xf3: // INK
      ink(mode);
      break;

    case 0xf4: // DEK
      dek(mode);
      break;

    case 0xf7: // LDX
      ldx(mode, arg);
      break;

    default:
      halt(true);
      throw Error::illegal_instruction(InstructionPointer);
    }

    if (!jump) {
      InstructionPointer += len;
    } else {
      jump = false;
    }
    InstructionPointer &= 0xffff;

    return (! halted);
  }

#ifndef GUI_ONLY
  void CPU::dump()
  {
    std::cerr.setf(std::ios::basefield, std::ios::hex);
    std::cerr.unsetf(std::ios::dec);
    //std::cerr.width(4);
    std::cerr.fill('0');
    std::cerr << "Akku:  " << std::setw(4) << (Akku & 0xffff) << std::endl;
    std::cerr << "Index: " << std::setw(4) << (Index & 0xffff) << std::endl
              << "IP:    " << std::setw(4) << InstructionPointer << std::endl
              << "SP:    " << std::setw(4) << StackPointer << std::endl;
  }
#endif

}
