/* $Id: CPU.hpp,v 1.9 2004/11/21 20:52:49 markus Exp $ */

/*
 * Copyright (c) 2003 Markus Hennecke
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the Author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * The declaration of the class CPU describing the hypothetical machinecode
 * architecture.
 * The cpu needs memory and a decoder to operate.
 */
#ifndef HASI_CPU_HPP
#define HASI_CPU_HPP

// system includes:
#ifndef GUI_ONLY
#include <iostream>
#endif

// program includes:
#include "MemoryChain.hpp"
#include "Error.hpp"
#include "Decoder.hpp"

namespace Interpreter {

  /*
   * class CPU
   */
  class CPU
  {
  private:
    // this one contains the whole memory
    MemoryChain &memChain;
    // the decoder
    // TODO: not yet implemented
    Decoder &decoder;
    
    // following are all the functions that implement one operation of the
    // cpu. These are called after the decoder has decoded the instruction.

    void hlt(const Decoder::AddressMode mode)
      throw (Error::illegal_instruction);
    void lad(const Decoder::AddressMode mode, const int arg)
      throw (Error::illegal_instruction);
    void spi(const Decoder::AddressMode mode, const int arg)
      throw (Error::illegal_instruction);
    void add(const Decoder::AddressMode mode, const int arg)
      throw (Error::illegal_instruction);
    void sub(const Decoder::AddressMode mode, const int arg)
      throw (Error::illegal_instruction);
    void mul(const Decoder::AddressMode mode, const int arg)
      throw (Error::illegal_instruction);
    void div(const Decoder::AddressMode mode, const int arg)
      throw (Error::illegal_instruction, Error::division_by_zero);
    // jump commands:
    void jmp(const Decoder::AddressMode mode, const int arg)
      throw (Error::illegal_instruction);
    void jez(const Decoder::AddressMode mode, const int arg)
      throw (Error::illegal_instruction);
    void jgz(const Decoder::AddressMode mode, const int arg)
      throw (Error::illegal_instruction);
    void jlz(const Decoder::AddressMode mode, const int arg)
      throw (Error::illegal_instruction);
    void swap(const Decoder::AddressMode mode)
      throw (Error::illegal_instruction);
    void jsr(const Decoder::AddressMode mode, const int arg)
      throw (Error::illegal_instruction);
    void ret(const Decoder::AddressMode mode)
      throw (Error::illegal_instruction);
    // index commands:
    void cli(const Decoder::AddressMode mode)
      throw (Error::illegal_instruction);
    void ldi(const Decoder::AddressMode mode)
      throw (Error::illegal_instruction);
    void lia(const Decoder::AddressMode mode)
      throw (Error::illegal_instruction);
    void ink(const Decoder::AddressMode mode)
      throw (Error::illegal_instruction);
    void dek(const Decoder::AddressMode mode)
      throw (Error::illegal_instruction);
    void ldx(const Decoder::AddressMode mode, const int arg)
      throw (Error::illegal_instruction);
    // stack commands:
    void lsa(const Decoder::AddressMode mode)
      throw (Error::illegal_instruction);
    void lds(const Decoder::AddressMode mode)
      throw (Error::illegal_instruction);
    void pushac(const Decoder::AddressMode mode)
      throw (Error::illegal_instruction);
    void popac(const Decoder::AddressMode mode)
      throw (Error::illegal_instruction);
    void addst(const Decoder::AddressMode mode)
      throw (Error::illegal_instruction);
    void subst(const Decoder::AddressMode mode)
      throw (Error::illegal_instruction);
    void mulst(const Decoder::AddressMode mode)
      throw (Error::illegal_instruction);
    void divst(const Decoder::AddressMode mode)
      throw (Error::illegal_instruction, Error::division_by_zero);
    void push(const Decoder::AddressMode mode, const int arg)
      throw (Error::illegal_instruction);
    void pop(const Decoder::AddressMode mode, const int arg)
      throw (Error::illegal_instruction);

    // following are the internal registers of the cpu:
    short Akku;
    short Help;
    unsigned int Index;
    unsigned int InstructionPointer;
    unsigned int StackPointer;

    // status of the cpu
    bool halted;

    // jump will be set true if a change of InstructionPointer
    // was made while executing the instruction
    bool jump;

    // len holds the length of the last decoded command
    int len;

    // lastWriteAccess holds the address of the last write
    unsigned int lastWriteAddress;

  public:
    // Con and Destructor:
    CPU(MemoryChain &chain, Decoder &decoder)
      : memChain(chain), decoder(decoder), Akku(0), Index(0),
        InstructionPointer(0), StackPointer(65535), halted(false),
        jump(false), len(0), lastWriteAddress(0)
    { }
    ~CPU() { };

    // get and set the registers:
    short getAkku() const { return Akku; }
    void  setAkku(const short newval) { Akku = newval; }
    short getHelp() const { return Help; }
    void  setHelp(const short newval) { Help = newval; };
    short getIndex() const { return Index; }
    void  setIndex(const short newval) { Index = newval; }
    unsigned short getStackPointer() const { return StackPointer; }
    void  setStackPointer(const unsigned short newval)
    { StackPointer = newval; }
    unsigned short getInstructionPointer() const { return InstructionPointer; }
    void  setInstructionPointer(const unsigned short newval) 
    { InstructionPointer = newval; }

    // get and set the status of the cpu
    bool isHalted() const { return halted; }
    void halt(const bool halt_now) { halted = halt_now; }

    // step the cpu. this is the main execution function.
    bool step() throw (Error::illegal_instruction, Error::division_by_zero);

    // get info about written memory
    unsigned int getLastWriteAddress() const { return lastWriteAddress; };

    // define the connected "Hardware"
    Decoder &getDecoder() { return decoder; };
    MemoryChain &getMemoryChain() { return memChain; };

#ifndef GUI_ONLY
    // dumps the contents of the cpu registers
    void dump();
#endif
  };

}

#endif
