/* $Id: Decoder.cpp,v 1.7 2003/12/12 09:47:27 markus Exp $ */

/*
 * Copyright (c) 2003 Markus Hennecke
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the Author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* The declaration of the Instruction decoder.
 * If the symbol DEBUG is defined the decoder is capable of returning strings
 * representing the instruction decoded.
 */

// program includes:
#include "Decoder.hpp"

// system includes:
#include <sstream>
#include <iomanip>
#include <map>

namespace Interpreter {

  // Decoder::Decoder(MemoryChain &memChain)
  //   : memChain(memChain)
  // Constructor. Initializes the map for disassembling.
  Decoder::Decoder(MemoryChain &memChain)
    : memChain(memChain)
  {
    // initialize the translation table
    mnemonics[0x0] = "HLT";
    mnemonics[0x1] = "LAD";
    mnemonics[0x2] = "SPI";
    mnemonics[0x3] = "ADD";
    mnemonics[0x4] = "SUB";
    mnemonics[0x5] = "MUL";
    mnemonics[0x6] = "DIV";
    mnemonics[0x7] = "JMP";
    mnemonics[0x8] = "JEZ";
    mnemonics[0x9] = "JGZ";
    mnemonics[0xa] = "JLZ";
    mnemonics[0xb] = "SWAP";
    mnemonics[0xd] = "JSR";
    mnemonics[0xe] = "RET";
    mnemonics[0xc1] = "PUSHAC";
    mnemonics[0xc2] = "POPAC";
    mnemonics[0xc3] = "ADDST";
    mnemonics[0xc4] = "SUBST";
    mnemonics[0xc5] = "MULST";
    mnemonics[0xc6] = "DIVST";
    mnemonics[0xc7] = "LSA";
    mnemonics[0xc8] = "LDS";
    mnemonics[0xcb] = "PUSH";
    mnemonics[0xcc] = "POP";
    mnemonics[0xf0] = "CLI";
    mnemonics[0xf1] = "LDI";
    mnemonics[0xf2] = "LIA";
    mnemonics[0xf3] = "INK";
    mnemonics[0xf4] = "DEK";
    mnemonics[0xf7] = "LDX";
  }

  // Decoder::~Decoder()
  // Nothing to do here...
  Decoder::~Decoder()
  { }

  // int Decoder::decode(const unsigned int address, AddressMode &mode,
  //                     int &code, int &arg)
  // Decodes the instruction that is at the given address in the memory
  // chain. The variables mode, code and arg are set accordingly to the
  // decoded address.
  int Decoder::decode(const unsigned int address, AddressMode &mode,
                      int &code, int &arg)
  {
    short read_val, read_arg;
    bool double_word = false;

    // read the instruction
    memChain.read(address, read_val);

    // isolate the opcode
    code = read_val & 0xf000;

    switch (code) {
    case 0xf000:
      // index operations...
      switch (read_val & 0x0700) {
      case 0x0700:
        // ldx @address // always double word
        memChain.read(address + 1, read_arg);
        arg = static_cast<int>(read_arg);
        double_word = true;
        code = 0xf7;
        mode = ADDRESS;
        break;

      case 0x0500:
        throw Error::illegal_instruction(address);

      default:
        code = (read_val >> 8) & 0x00f7;
        mode = NONE;
        break;
      }
      break;

    case 0xc000:
      // stack operations...
      switch (read_val & 0x0f00) {
      case 0x0000:
      case 0x0900:
      case 0x0a00:
      case 0x0d00:
      case 0x0e00:
      case 0x0f00:        
        throw Error::illegal_instruction(address);

      case 0x0b00:
        // PUSH
        mode = DIRECT;
        code = 0xcb;
        goto read16bitarg;
        
      case 0x0c00:
        mode = ADDRESS;
        code = 0xcc;
      read16bitarg:
        memChain.read(address + 1, read_arg);
        arg = static_cast<int>(read_arg);
        double_word = true;
        break;
      default:
        mode = NONE;
        code = (read_val >> 8) & 0xff;
        break;
      }
      break;

      case 0xb000:        
        // former illegal opcode
        // throw Error::illegal_instruction(address);

        // this is treated special for now. the swap instruction has
        // this opcode
        mode = NONE;
        code = 0xb;
        break;

    default:
      // this is the default case
      // check for a double word command and read the arg if necessary
      if (read_val & 0x0800) {
        memChain.read(address + 1, read_arg);
        double_word = true;
      }

      // a opcode with a normal mod part
      switch (read_val & 0x0700) {
      case 0x0600:
        // indirect: (xxxx)
        mode = INDIRECT;
        if (! double_word)
          read_arg = read_val & 0x00ff;
        break;

      case 0x0700:
        // index: %xxxx
        mode = INDEX;
        goto read8bitarg;

      case 0x0500:
        // relative: xxxx
        mode = RELATIVE;
        goto read8bitarg;
        
      case 0x0400:
        // direct operand: #xxxx
        mode = DIRECT;
      read8bitarg:
        if (! double_word) {
          read_arg = read_val & 0x00ff;
          // adjust sign
          if (read_arg > 0x007f)
            read_arg |= 0xff00;
        }
        break;

      default:
        // direct address: @xxxx
        if (! double_word) 
          read_arg = read_val & 0x03ff;
        mode = ADDRESS;
        break;
      }
      code = (read_val >> 12) & 0x000f;

      // check for a single word instruction with no addresses...
      if ((code == 0x0) || (code == 0xe)) {
        mode = NONE;
      }
    }

    // make the arg a short...
    arg = static_cast<int>(read_arg);
    arg &= 0xffff;

    // return the number of words decoded
    return (double_word ? 2 : 1);
  }

  // int Decoder::decode(const unsigned int address, std::string &addr,
  //                     std::string &words, std::string &opcode, 
  //                     std::string &arg)
  // Decodes the opcode at the given address and disassembles it to the
  // strings addr, words, opcode and arg.
  int Decoder::decode(const unsigned int address, std::string &addr,
                      std::string &words, std::string &opcode, 
                      std::string &arg)
  {
    int dec_arg, dec_code, len;
    AddressMode mode;
    
    std::ostringstream ost;

    ost.fill('0');
    ost << std::hex << std::setw(4) << (address & 0xffff) << ":";
    addr = ost.str();
    ost.str("");

    try {
      // get the mode, code and argument.
      // if it is an illegal instruction we will raise an exception but
      // addr is already set to valid values
      len = decode(address, mode, dec_code, dec_arg);

      if (mnemonics.find(dec_code) != mnemonics.end()) {
        // ok, this seems to be an valid opcode
        opcode = mnemonics[dec_code];
      } else {
        throw Error::illegal_instruction(address);
      }

      switch (mode) {
      default:
        // NONE...
        if ((dec_code == 0x01) || (dec_code == 0x02) || (dec_code == 0x03) ||
            (dec_code == 0x04) || (dec_code == 0x05) || (dec_code == 0x06) ||
            (dec_code == 0x07) || (dec_code == 0x08) || (dec_code == 0x09) ||
            (dec_code == 0x0d) || (dec_code == 0xcb) || (dec_code == 0xcc) ||
            (dec_code == 0xf7))
          throw Error::illegal_instruction(address);
        break;
      case DIRECT:
        if ((dec_code != 0x01) && (dec_code != 0x03) && (dec_code != 0x04) &&
            (dec_code != 0x05) && (dec_code != 0x06) && (dec_code != 0xcb))
          throw Error::illegal_instruction(address);
        ost << "#" << std::hex << std::setw(4) << (dec_arg & 0xffff);
        break;
      case ADDRESS:
        if ((dec_code != 0x01) && (dec_code != 0x02) && (dec_code != 0x03) &&
            (dec_code != 0x04) && (dec_code != 0x05) && (dec_code != 0x06) &&
            (dec_code != 0x07) && (dec_code != 0x08) && (dec_code != 0x09) &&
            (dec_code != 0x0a) && (dec_code != 0x0d) && (dec_code != 0xf7) &&
            (dec_code != 0xcc))
          throw Error::illegal_instruction(address);
        ost << "@" << std::hex << std::setw(4) << (dec_arg & 0xffff);
        break;
      case INDEX:
        if ((dec_code != 0x01) && (dec_code != 0x02) && (dec_code != 0x03) &&
            (dec_code != 0x04) && (dec_code != 0x05) && (dec_code != 0x06) &&
            (dec_code != 0x07) && (dec_code != 0x08) && (dec_code != 0x09) &&
            (dec_code != 0x0a) && (dec_code != 0x0d))
          throw Error::illegal_instruction(address);
        ost << "%" << std::hex << std::setw(4) << (dec_arg & 0xffff);
        break;
      case RELATIVE:
        if ((dec_code != 0x01) && (dec_code != 0x02) && (dec_code != 0x03) &&
            (dec_code != 0x04) && (dec_code != 0x05) && (dec_code != 0x06) &&
            (dec_code != 0x07) && (dec_code != 0x08) && (dec_code != 0x09) &&
            (dec_code != 0x0a) && (dec_code != 0x0d))
          throw Error::illegal_instruction(address);
        ost << std::hex << std::setw(4) << (dec_arg & 0xffff);
        break;
      case INDIRECT:
        if ((dec_code != 0x01) && (dec_code != 0x02) && (dec_code != 0x03) &&
            (dec_code != 0x04) && (dec_code != 0x05) && (dec_code != 0x06) &&
            (dec_code != 0x07) && (dec_code != 0x08) && (dec_code != 0x09) &&
            (dec_code != 0x0a) && (dec_code != 0x0d))
          throw Error::illegal_instruction(address);
        ost << "(" << std::hex << std::setw(4) << (dec_arg & 0xffff) << ")";
        break;
      }
      arg = ost.str();
    }
    catch (Error::illegal_instruction) {
      arg = "";
      opcode = "Illegal Instruction";
      len = 1;
    }

    ost.str("");
    ost.fill('0');
    short mem;
    if (memChain.read(address, mem)) {
      ost << std::hex << std::setw(4) << (mem & 0xffff) << " ";
      if (len > 1) {
        memChain.read(address + 1, mem);
        ost << std::hex << std::setw(4) << (mem & 0xffff);
      } else {
        ost << "    ";
      }
    } else {
      ost << "---- ----";
    }
    words = ost.str();

    return len;
  }

}
