/* $Id: ObjectFile.cpp,v 1.5 2003/12/12 09:47:27 markus Exp $ */

/*
 * Copyright (c) 2003 Markus Hennecke
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the Author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Implementation of an class handling object files.
 */

// program includes:
#include "ObjectFile.hpp"

namespace Interpreter {

  // ObjectFile::ObjectFile(const std::string &)
  // Constructor. Creates an input stream for the file specified by filename.
  ObjectFile::ObjectFile(const std::string &filename)
    throw (Error::file_open_error)
    : filename(filename)
  {
    file = new std::ifstream(filename.c_str());
    if (! file->is_open())
      throw Error::file_open_error(filename);
  }

  // ObjectFile::~ObjectFile()
  // Destructor. Closes the file if it was open.
  ObjectFile::~ObjectFile()
  {
    if (file)
      delete file;
  }

  // void ObjectFile::splitLine(const std::string &)
  // Split the line. The first string will be the address, each following
  // string will be a word. If the line does not fit the object file format
  // false will be returned.
  bool ObjectFile::splitLine(const std::string &str)
  {
    std::string s(str);
    // first clear the queue
    stringparts.clear();

    std::string::size_type marker = s.find(":");
    if ((s.length() > 0) && (marker == std::string::npos))
      // the colon after the address was not found...
      return false;

    stringparts.push_back(s.substr(0, marker));
    s.erase(0, ++marker);
 
    while (! s.empty()) {
      // remove leading whitespaces
      while ((s.length() > 0) && (isspace(s[0])))
        s.erase(0, 1);

      if (s.empty())
        // stop if there are no more characters
        break;

      marker = s.find(" ");
      if (marker == std::string::npos)
        // the whole string 
        marker = s.length();

      stringparts.push_back(s.substr(0, marker));
      s.erase(0, marker);
    }

    return true;
  }

  // void ObjectFile::removeComment(std::string &)
  // Removes a possible comment at the end of the string.
  void ObjectFile::removeComment(std::string &s)
  {
    std::string::size_type semicolon = s.rfind(";");
    if (semicolon == std::string::npos)
      // not found, nothing to do
      return;

    s.erase(semicolon, s.length() - 1);
  }
  
  // int ObjectFile::hexStringToInt(const std::string &)
  // Converts a 4 digits hex string to a number. Error::file_read_error will
  // be thrown if the string is not valid.
  int ObjectFile::hexStringToInt(const std::string &s)
    throw (Error::file_read_error)
  {
    int val = 0;

    // check if the size is correct
    if (s.length() != 4)
      throw Error::file_read_error(filename);

    for (int i = 0; i < 4; i++) {
      if (! isxdigit(s[i]))
        throw Error::file_read_error(filename);
      val *= 16;
      int asciival = toupper(s[i]) - 0x30;
      val += (asciival > 9) ? asciival - 7 : asciival;
    }

    return val;
  }

  // void ObjectFile::read(MemoryChain *) throw (Error::file_read_error)
  // Reads an object file and stores the read data into the virtual memory.
  void ObjectFile::read(MemoryChain *mem) throw (Error::file_read_error)
  {
    std::string line;
    while (! file->eof()) {
      std::getline(*file, line);
#ifdef DEBUG
      std::cerr << line << std::endl;
#endif
 
      // remove comments
      removeComment(line);

      // skip empty lines
      if (line.empty())
        continue;

      if (splitLine(line)) {
        unsigned int address = 0;
        for (std::list<std::string>::const_iterator it = stringparts.begin();
             it != stringparts.end(); it++) {
          if (it == stringparts.begin()) {
            address = static_cast<unsigned int>(hexStringToInt(*it));
            continue;
          }
          mem->write(address++, hexStringToInt(*it));
        }
      }
    }
  }

}
