/* $Id: main.cpp,v 1.11 2004/11/21 20:54:24 markus Exp $ */

/*
 * Copyright (c) 2003 Markus Hennecke
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the Author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * The main program...
 */

// system includes:
#include <string>
#include <iostream>
#include <iomanip>

// program includes:
#include "MemoryChain.hpp"
#include "Memory.hpp"
#include "RAM.hpp"
#include "IOTtyOut.hpp"
#include "CPU.hpp"
#include "Decoder.hpp"
#include "ObjectFile.hpp"

unsigned int start = 0;
bool dump_regs = false;
std::string objfile;

#define VERSION "Version 1.1-pre2"

// static void usage()
// Prints out how to use the program and exits with an exit code of 1.
static void usage()
{
  std::cout << "HASI - Hypothetical Machinecode Assembler and Simulator"
            << std::endl << std::endl
            << "hasi-nogui - HASI standalone interpreter" << std::endl;
  std::cout << "Copyright (c) 2003 Markus Hennecke" << std::endl
            << VERSION << std::endl << std::endl;

  std::cout << "USAGE: hasi-nogui [options] objectfile" << std::endl;
  std::cout << "\t-h           this help" << std::endl;
  std::cout << "\t-d           generate debug output (registers)" << std::endl;
  std::cout << "\t-s address   set the start address" << std::endl;
 
  exit(1);
}

// static void parse_commandline(int argc, char **argv)
// Parses the commandline arguments. If something is not correct the usage
// information is printed and the program is terminated with an exit code 
// of 1.
static void parse_commandline(int argc, char **argv)
{
  for (int i = 1; i < argc; i++) {
    std::string opt(argv[i]);
    if (opt == "-h") {
      usage();
    } else if (opt == "-d") {
      dump_regs = true;
    } else if (opt == "-s") {
      if (++i >= argc)
        usage();
      char *endptr = NULL;
      start = strtoul(argv[i], &endptr, 10);
      if (*endptr != '\0')
        usage();
      if (start > 65535) 
        usage();
    } else {
      objfile = argv[i++];
      if (! objfile.length()) {
        usage();
      }
      if (i != argc)
        usage();
    }
  }
  
  // check if there was an object file...
  if (objfile.empty())
    usage();
}

int main(int argc, char **argv)
{
  parse_commandline(argc, argv);

  try {
    // Possible TODO: create Memory user defined??
    // This would include reading a config file...
    Interpreter::MemoryChain memChain;
    Interpreter::RAM *mem = new Interpreter::RAM(65536);
    memChain.addMemorySegment(mem);
    Interpreter::IOTtyOut *outport = new Interpreter::IOTtyOut(0xc000);
    memChain.addMemorySegment(outport);
    Interpreter::Decoder decoder(memChain);

    Interpreter::CPU cpu(memChain, decoder); 
    cpu.setInstructionPointer(start);

    Interpreter::ObjectFile *ofile = new Interpreter::ObjectFile(objfile);
    ofile->read(&memChain);

    // do the cpu, yeah!

    while (! cpu.isHalted()) {

      if (dump_regs) {
	std::string addr, words, opcode, arg;
	decoder.decode(cpu.getInstructionPointer(), addr, words, opcode, arg);
	std::cerr << addr << " " << words << " " << opcode << " " << arg
		  << std::endl;
        cpu.dump();
      }

      cpu.step();
    }  
    delete mem;
    delete outport;
  }
  catch (Error::division_by_zero &div) {
    std::cerr << "Error: division by zero at ";
    std::cerr.setf(std::ios::basefield, std::ios::hex);
    std::cerr.unsetf(std::ios::dec);
    std::cerr.fill('0');
    std::cerr << std::setw(4) << div.getAddress() << std::endl;
  }

  catch (Error::illegal_instruction &illegal) {
    std::cerr << "Error: illegal instruction at ";
    std::cerr.setf(std::ios::basefield, std::ios::hex);
    std::cerr.unsetf(std::ios::dec);
    std::cerr.fill('0');
    std::cerr << std::setw(4) << illegal.getAddress() << std::endl;
  }

  catch (Error::file_open_error &e) {
    e.emit();
  }
  catch (Error::file_read_error &e) {
    e.emit();
  }

  return 0;
}
